# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import logging

import faebryk.library._F as F
from faebryk.core.module import Module
from faebryk.libs.library import L
from faebryk.libs.units import P

logger = logging.getLogger(__name__)


# TODO: make generic (use Switch module, different switch models, bistable, etc.)
class Relay(Module):
    switch_a_nc: F.Electrical
    switch_a_common: F.Electrical
    switch_a_no: F.Electrical
    switch_b_no: F.Electrical
    switch_b_common: F.Electrical
    switch_b_nc: F.Electrical
    coil_power: F.ElectricPower

    coil_max_voltage = L.p_field(units=P.V)
    coil_max_current = L.p_field(units=P.A)
    coil_resistance = L.p_field(units=P.ohm)
    contact_max_switching_voltage = L.p_field(units=P.V)
    contact_max_switching_current = L.p_field(units=P.A)
    contact_max_current = L.p_field(units=P.A)

    designator_prefix = L.f_field(F.has_designator_prefix)(
        F.has_designator_prefix.Prefix.K
    )

    usage_example = L.f_field(F.has_usage_example)(
        example="""
        import Relay, ElectricPower, Diode, NFET, ElectricLogic

        relay = new Relay
        relay.coil_max_voltage = 12V
        relay.coil_max_current = 50mA
        relay.coil_resistance = 240ohm +/- 10%
        relay.contact_max_switching_voltage = 250V
        relay.contact_max_switching_current = 10A
        relay.contact_max_current = 16A

        # Connect coil power
        power_12v = new ElectricPower
        assert power_12v.voltage within 12V +/- 5%
        relay.coil_power ~ power_12v

        # Control relay with MOSFET and flyback diode
        control_mosfet = new NFET
        flyback_diode = new Diode
        control_signal = new ElectricLogic

        # Coil control circuit
        power_12v.hv ~ relay.coil_power.hv
        relay.coil_power.lv ~> control_mosfet ~> power_12v.lv
        control_signal ~ control_mosfet.gate

        # Flyback diode for coil protection
        relay.coil_power.hv ~ flyback_diode.cathode
        flyback_diode.anode ~ relay.coil_power.lv

        # Switch high-power load using normally open contact
        high_power_load ~ relay.switch_a_no
        relay.switch_a_common ~ high_voltage_supply
        """,
        language=F.has_usage_example.Language.ato,
    )



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import faebryk.library._F as F
from faebryk.core.module import Module
from faebryk.libs.library import L
from faebryk.libs.units import P


class Crystal(Module):
    # ----------------------------------------
    #     modules, interfaces, parameters
    # ----------------------------------------
    gnd: F.Electrical
    unnamed = L.list_field(2, F.Electrical)

    # ----------------------------------------
    #               parameters
    # ----------------------------------------
    frequency = L.p_field(
        units=P.Hz,
        likely_constrained=True,
        soft_set=L.Range(32.768 * P.kHz, 100 * P.MHz),
        tolerance_guess=50 * P.ppm,
    )

    frequency_tolerance = L.p_field(
        units=P.ppm,
        likely_constrained=True,
        soft_set=L.Range(10 * P.ppm, 100 * P.ppm),
        tolerance_guess=10 * P.percent,
    )

    frequency_temperature_tolerance = L.p_field(
        units=P.ppm,
        likely_constrained=True,
        soft_set=L.Range(1 * P.ppm, 50 * P.ppm),
        tolerance_guess=10 * P.percent,
    )

    frequency_ageing = L.p_field(
        units=P.ppm,
        likely_constrained=True,
        soft_set=L.Range(1 * P.ppm, 10 * P.ppm),
        tolerance_guess=20 * P.percent,
    )

    equivalent_series_resistance = L.p_field(
        units=P.Ω,
        likely_constrained=True,
        soft_set=L.Range(10 * P.Ω, 200 * P.Ω),
        tolerance_guess=10 * P.percent,
    )

    shunt_capacitance = L.p_field(
        units=P.F,
        likely_constrained=True,
        soft_set=L.Range(1 * P.pF, 10 * P.pF),
        tolerance_guess=20 * P.percent,
    )

    load_capacitance = L.p_field(
        units=P.F,
        likely_constrained=True,
        soft_set=L.Range(8 * P.pF, 30 * P.pF),
        tolerance_guess=10 * P.percent,
    )

    # ----------------------------------------
    #                traits
    # ----------------------------------------
    designator = L.f_field(F.has_designator_prefix)(F.has_designator_prefix.Prefix.XTAL)
    footprint: F.can_attach_to_footprint_symmetrically

    # ----------------------------------------
    #                connections
    # ----------------------------------------

    usage_example = L.f_field(F.has_usage_example)(
        example="""
        import Crystal, Capacitor

        crystal = new Crystal
        crystal.frequency = 16MHz +/- 20ppm
        crystal.frequency_tolerance = 20ppm
        crystal.load_capacitance = 18pF +/- 10%
        crystal.equivalent_series_resistance = 80ohm +/- 20%
        crystal.package = "HC49U"

        # Connect to microcontroller with load capacitors
        load_cap1 = new Capacitor
        load_cap2 = new Capacitor
        load_cap1.capacitance = 22pF +/- 5%
        load_cap2.capacitance = 22pF +/- 5%

        mcu.xtal1 ~ crystal.unnamed[0]
        mcu.xtal2 ~ crystal.unnamed[1]
        crystal.unnamed[0] ~> load_cap1 ~> crystal.gnd
        crystal.unnamed[1] ~> load_cap2 ~> crystal.gnd
        crystal.gnd ~ power_supply.lv
        """,
        language=F.has_usage_example.Language.ato,
    )



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import faebryk.library._F as F  # noqa: F401
from faebryk.core.module import Module  # noqa: F401
from faebryk.libs.library import L  # noqa: F401


class LEDIndicator(Module):
    # interfaces

    logic_in: F.ElectricLogic
    power_in: F.ElectricPower

    # components

    led: F.PoweredLED

    def __init__(self, use_mosfet: bool = False, active_low: bool = False):
        super().__init__()
        self._use_mosfet = use_mosfet
        self._active_low = active_low

    def __preinit__(self):
        if self._use_mosfet:
            power_switch = self.add(
                F.PowerSwitchMOSFET(lowside=True, normally_closed=self._active_low)
            )
        else:
            power_switch = self.add(
                F.PowerSwitchStatic(normally_closed=self._active_low)
            )

        self.power_in.connect_via(power_switch, self.led.power)
        power_switch.logic_in.connect(self.logic_in)

    usage_example = L.f_field(F.has_usage_example)(
        example="""
        import LEDIndicator, ElectricPower, ElectricLogic

        # Basic LED indicator (active high)
        status_led = new LEDIndicator(use_mosfet=False, active_low=False)

        # Connect power supply
        power_5v = new ElectricPower
        assert power_5v.voltage within 5V +/- 5%
        status_led.power_in ~ power_5v

        # Connect control signal
        status_signal = new ElectricLogic
        status_signal.reference ~ power_5v
        status_led.logic_in ~ status_signal

        # Connect to microcontroller
        microcontroller.gpio_status ~ status_signal.line

        # Configure LED properties
        status_led.led.led.color = LED.Color.GREEN
        status_led.led.led.current = 20mA +/- 5%
        status_led.led.led.forward_voltage = 2.1V +/- 10%

        # High-power LED with MOSFET driver
        power_led = new LEDIndicator(use_mosfet=True, active_low=False)
        power_led.power_in ~ power_5v
        power_led.logic_in ~ power_control_signal
        power_led.led.led.current = 100mA +/- 10%  # Higher current

        # Active-low indicator (ON when signal is LOW)
        error_led = new LEDIndicator(use_mosfet=False, active_low=True)
        error_led.power_in ~ power_5v
        error_led.logic_in ~ error_signal
        error_led.led.led.color = LED.Color.RED

        # Common applications: status indication, debugging, user feedback
        """,
        language=F.has_usage_example.Language.ato,
    )



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import logging
from enum import Enum, auto

import faebryk.library._F as F
from faebryk.core.module import Module
from faebryk.libs.library import L
from faebryk.libs.units import P

logger = logging.getLogger(__name__)


class Fuse(Module):
    class FuseType(Enum):
        NON_RESETTABLE = auto()
        RESETTABLE = auto()

    class ResponseType(Enum):
        SLOW = auto()
        FAST = auto()

    unnamed = L.list_field(2, F.Electrical)
    fuse_type = L.p_field(
        domain=L.Domains.ENUM(FuseType),
    )
    response_type = L.p_field(
        domain=L.Domains.ENUM(ResponseType),
    )
    trip_current = L.p_field(
        units=P.A,
        likely_constrained=True,
        domain=L.Domains.Numbers.REAL(),
        soft_set=L.Range(100 * P.mA, 100 * P.A),
    )

    attach_to_footprint: F.can_attach_to_footprint_symmetrically

    @L.rt_field
    def can_bridge(self):
        return F.can_bridge_defined(self.unnamed[0], self.unnamed[1])

    designator_prefix = L.f_field(F.has_designator_prefix)(
        F.has_designator_prefix.Prefix.F
    )

    usage_example = L.f_field(F.has_usage_example)(
        example="""
        import Fuse, ElectricPower

        fuse = new Fuse
        fuse.trip_current = 2A +/- 10%
        fuse.fuse_type = Fuse.FuseType.NON_RESETTABLE
        fuse.response_type = Fuse.ResponseType.FAST
        fuse.package = "1206"

        # Connect fuse in series with power supply
        power_input = new ElectricPower
        protected_power = new ElectricPower

        # Fuse protects the circuit from overcurrent
        power_input.hv ~> fuse ~> protected_power.hv
        power_input.lv ~ protected_power.lv

        # For resettable fuse (PTC)
        ptc_fuse = new Fuse
        ptc_fuse.trip_current = 500mA +/- 20%
        ptc_fuse.fuse_type = Fuse.FuseType.RESETTABLE
        ptc_fuse.response_type = Fuse.ResponseType.SLOW

        # Common applications: USB power protection, battery protection
        """,
        language=F.has_usage_example.Language.ato,
    )



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import faebryk.library._F as F
from faebryk.core.module import Module
from faebryk.libs.library import L
from faebryk.libs.units import P


class EEPROM(Module):
    """
    Generic EEPROM module with F.I2C interface.
    """

    def set_address(self, addr: int):
        """
        Configure the address of the EEPROM by setting the address pins.
        """
        assert addr < (1 << len(self.address))

        for i, e in enumerate(self.address):
            e.set(addr & (1 << i) != 0)

    # ----------------------------------------
    #     modules, interfaces, parameters
    # ----------------------------------------

    memory_size = L.p_field(
        units=P.bit,
        likely_constrained=True,
        domain=L.Domains.Numbers.NATURAL(),
        soft_set=L.Range(128 * P.bit, 1024 * P.kbit),
    )

    power: F.ElectricPower
    i2c: F.I2C
    write_protect: F.ElectricLogic
    address = L.list_field(3, F.ElectricLogic)

    # ----------------------------------------
    #                traits
    # ----------------------------------------

    designator_prefix = L.f_field(F.has_designator_prefix)(
        F.has_designator_prefix.Prefix.U
    )

    @L.rt_field
    def single_electric_reference(self):
        return F.has_single_electric_reference_defined(
            F.ElectricLogic.connect_all_module_references(self)
        )

    usage_example = L.f_field(F.has_usage_example)(
        example="""
        import EEPROM, ElectricPower, I2C, Resistor

        eeprom = new EEPROM
        eeprom.memory_size = 32kbit

        # Connect power supply
        power_3v3 = new ElectricPower
        assert power_3v3.voltage within 3.3V +/- 5%
        eeprom.power ~ power_3v3

        # Connect I2C bus
        i2c_bus = new I2C
        i2c_bus.frequency = 400kHz  # Fast mode
        eeprom.i2c ~ i2c_bus

        # Connect to microcontroller
        microcontroller.i2c ~ i2c_bus

        # Set device address using address pins
        eeprom.set_address(0)  # Device address 0b000

        # Connect address pins to power rails for different addresses
        eeprom.address[0].line ~ power_3v3.lv  # A0 = 0
        eeprom.address[1].line ~ power_3v3.lv  # A1 = 0
        eeprom.address[2].line ~ power_3v3.lv  # A2 = 0

        # Write protect control (optional)
        eeprom.write_protect.reference ~ power_3v3
        eeprom.write_protect.line ~ power_3v3.lv  # Enable writes

        # Pull-up resistors for I2C (if not provided elsewhere)
        sda_pullup = new Resistor
        scl_pullup = new Resistor
        sda_pullup.resistance = 4.7kohm +/- 5%
        scl_pullup.resistance = 4.7kohm +/- 5%
        i2c_bus.sda.line ~> sda_pullup ~> power_3v3.hv
        i2c_bus.scl.line ~> scl_pullup ~> power_3v3.hv
        """,
        language=F.has_usage_example.Language.ato,
    )



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

from enum import StrEnum

import faebryk.library._F as F
import faebryk.libs.library.L as L
from faebryk.core.parameter import Parameter


class is_pickable_by_type(F.is_pickable.decless()):
    """
    Marks a module as being parametrically selectable using the given parameters.

    Must map to an existing API endpoint.

    Should be named "pickable" to aid overriding by subclasses.
    """

    class Endpoint(StrEnum):
        """Query endpoints known to the API."""

        RESISTORS = "resistors"
        CAPACITORS = "capacitors"
        INDUCTORS = "inductors"

    def __init__(self, endpoint: Endpoint, params: list[Parameter]):
        super().__init__()
        self.endpoint = endpoint
        self._params = params

    @property
    def params(self) -> list[Parameter]:
        return self._params

    @property
    def pick_type(self) -> type[L.Module]:
        return type(self.get_obj(L.Module))



"""
Configure the user's system for atopile development.
"""

import logging

from atopile.telemetry import capture

logger = logging.getLogger(__name__)


@capture("cli:configure_start", "cli:configure_end")
def configure() -> None:
    """
    Configure the user's system for atopile development.
    """
    from textwrap import dedent

    from faebryk.libs.logging import rich_print_robust

    # Just here for legacy support
    rich_print_robust(
        dedent(
            """
            This command is deprecated and will be removed in a future version.
            Configuration/Setup should be automatically handled.
            """
        )
    )


def setup() -> None:
    # Cleanup legacy config file
    from faebryk.libs.kicad.ipc import enable_plugin_api
    from faebryk.libs.paths import get_config_dir

    try:
        _LEGACY_CFG_PATH = get_config_dir() / "configured_for.yaml"
        if _LEGACY_CFG_PATH.exists():
            _LEGACY_CFG_PATH.unlink()
    except Exception as e:
        logger.warning(f"Couldn't remove legacy config file: {e!r}")

    try:
        install_kicad_plugin()
    except Exception as e:
        logger.warning(f"Couldn't install plugin: {e!r}")

    try:
        enable_plugin_api()
    except Exception as e:
        logger.warning(f"Couldn't enable plugin api: {e!r}")


def install_kicad_plugin() -> None:
    """Install the kicad plugin."""
    # TODO switch to new plugin as soon as group serialize ipc works in kicad
    # TODO then also remove legacy plugin from existing installations
    from atopile.kicad_plugin.lib import install_kicad_legacy_plugin

    install_kicad_legacy_plugin()



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT
import logging

import faebryk.library._F as F
from faebryk.core.moduleinterface import ModuleInterface
from faebryk.libs.library import L

logger = logging.getLogger(__name__)


class Ethernet(ModuleInterface):
    """
    1000BASE-T Gigabit Ethernet Interface
    """

    # Ethernet pairs
    pairs = L.list_field(4, F.DifferentialPair)

    # Status LEDs
    led_speed: F.ElectricLogic  # Speed LED
    led_link: F.ElectricLogic  # Link LED

    @L.rt_field
    def single_electric_reference(self):
        return F.has_single_electric_reference_defined(
            F.ElectricLogic.connect_all_module_references(self)
        )

    def __postinit__(self, *args, **kwargs):
        super().__postinit__(*args, **kwargs)
        self.led_speed.line.add(
            F.has_net_name("ETH_LED_SPEED", level=F.has_net_name.Level.SUGGESTED)
        )
        self.led_link.line.add(
            F.has_net_name("ETH_LED_LINK", level=F.has_net_name.Level.SUGGESTED)
        )
        for i, pair in enumerate(self.pairs):
            pair.p.line.add(
                F.has_net_name(f"ETH_P{i}", level=F.has_net_name.Level.SUGGESTED)
            )
            pair.n.line.add(
                F.has_net_name(f"ETH_P{i}", level=F.has_net_name.Level.SUGGESTED)
            )

    usage_example = L.f_field(F.has_usage_example)(
        example="""
        import Ethernet, ElectricPower

        ethernet = new Ethernet

        # Connect power reference for logic levels
        power_3v3 = new ElectricPower
        ethernet.led_speed.reference ~ power_3v3
        ethernet.led_link.reference ~ power_3v3

        # Connect to PHY or connector
        # Four differential pairs for 1000BASE-T
        ethernet_connector.tx_pairs[0] ~ ethernet.pairs[0]
        ethernet_connector.tx_pairs[1] ~ ethernet.pairs[1]
        ethernet_connector.rx_pairs[2] ~ ethernet.pairs[2]
        ethernet_connector.rx_pairs[3] ~ ethernet.pairs[3]

        # Connect status LEDs
        ethernet.led_speed ~ speed_led_output
        ethernet.led_link ~ link_led_output
        """,
        language=F.has_usage_example.Language.ato,
    )



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import faebryk.library._F as F
from faebryk.core.module import Module
from faebryk.libs.library import L


class Regulator(Module):
    power_in: F.ElectricPower
    power_out: F.ElectricPower

    def __preinit__(self):
        self.power_out.add(F.Power.is_power_source.impl()())
        self.power_in.add(F.Power.is_power_sink.impl()())

    @L.rt_field
    def can_bridge(self):
        return F.can_bridge_defined(self.power_in, self.power_out)

    usage_example = L.f_field(F.has_usage_example)(
        example="""
        import Regulator, ElectricPower

        regulator = new Regulator

        # Connect input power (unregulated)
        power_input = new ElectricPower
        assert power_input.voltage within 7V to 20V  # Wide input range
        regulator.power_in ~ power_input

        # Connect output power (regulated)
        power_output = new ElectricPower
        regulator.power_out ~ power_output

        # Output voltage depends on regulator type:
        # - LDO: Vout = Vin - Dropout_voltage
        # - Switching: Vout = Function(Vin, feedback network)

        # Connect to load
        load_circuit ~ power_output

        # Note: This is a generic regulator interface
        # Use specific regulator types for actual implementations:
        # - LDO for low noise, low efficiency
        # - Buck for high efficiency step-down
        # - Boost for step-up conversion
        # - Buck-boost for bidirectional conversion
        """,
        language=F.has_usage_example.Language.ato,
    )



import contextlib
import logging
import os
import subprocess
from pathlib import Path
from typing import Optional

import wx as wx_module  # type: ignore

# ATTENTION: RUNS IN PYTHON3.8

LOG_FILE = (Path(__file__).parent / "log.log").expanduser().absolute()
LOG_FILE.parent.mkdir(parents=True, exist_ok=True)
formatter = logging.Formatter("%(asctime)s - %(message)s", datefmt="%Y-%m-%d %H:%M:%S")
file_handler = logging.FileHandler(str(LOG_FILE), "w", "utf-8")
file_handler.setFormatter(formatter)


def setup_logger(module_name: str):
    log = logging.getLogger(module_name)
    log.addHandler(file_handler)
    log.setLevel(logging.INFO)
    return log


log = setup_logger(__name__)


def run_ato(args: list[str], cwd: Optional[Path] = None) -> subprocess.CompletedProcess:
    """Run the ato command with the given arguments."""
    executable = os.environ.get("ATOPILE_PYTHON")
    if not executable:
        raise RuntimeError("ATOPILE_PYTHON not set")
    cmd = [executable, "-m", "atopile"] + args
    env = os.environ.copy()
    env.pop("PYTHONPATH", None)
    env.pop("PYTHONHOME", None)
    env.pop("PYTHONEXECUTABLE", None)
    try:
        out = subprocess.run(
            cmd, check=True, text=True, cwd=cwd, capture_output=True, env=env
        )
    except FileNotFoundError:
        log.error(f"Could not find Python executable: {executable}")

        message_box(
            "Could not find Python interpreter or atopile module. "
            "Please ensure atopile is installed and run 'ato configure' "
            "to set up the KiCad plugin environment.",
            "Pull Group Error",
            wx_module.OK | wx_module.ICON_ERROR,
        )
        raise

    log.info(f"Ran ato command: {cmd}")
    log.info(f"stdo: {out.stdout}")
    log.info(f"stde: {out.stderr}")
    return out


def message_box(message: str, title: str, icon: int):
    """Show a message box."""
    wx = wx_module.GetApp()
    if wx:
        wx_module.MessageBox(message, title, icon)
    else:
        log.warning("No wx application found")


@contextlib.contextmanager
def log_exceptions():
    try:
        yield
    except Exception as e:
        log.exception(e)
        raise



# pylint: disable=logging-fstring-interpolation

"""
`ato inspect`
"""

import logging
from typing import Annotated

import typer

log = logging.getLogger(__name__)


def inspect(
    entry: Annotated[str | None, typer.Argument()] = None,
    build: Annotated[list[str], typer.Option("--build", "-b", envvar="ATO_BUILD")] = [],
    target: Annotated[
        list[str], typer.Option("--target", "-t", envvar="ATO_TARGET")
    ] = [],
    option: Annotated[
        list[str], typer.Option("--option", "-o", envvar="ATO_OPTION")
    ] = [],
    inspect: str | None = None,
    context: Annotated[
        str | None,
        typer.Option(
            "--context", "-c", help="The context from which to inspect the module"
        ),
    ] = None,
    dump_csv: Annotated[
        str | None,
        typer.Option("--dump-csv", "-d", help="Output the inspection to a CSV file"),
    ] = None,
):
    """
    Utility to inspect what is connected to a component.
    The context sets the boundary where something is considered connected.
    For example: `--inspect rp2040_micro --context rp2040_micro_ki`
    """
    from atopile import errors

    raise errors.UserNotImplementedError("Inspect is not yet implemented.")



import logging
from dataclasses import dataclass
from pathlib import Path
from textwrap import dedent
from typing import Literal

from dataclasses_json import DataClassJsonMixin, LetterCase, dataclass_json

from faebryk.libs.kicad.paths import get_plugin_paths

logger = logging.getLogger(__name__)

RESOURCE_DIR = Path(__file__).parent / "resource"


@dataclass_json(letter_case=LetterCase.KEBAB)  # type: ignore
@dataclass
class KicadPluginManifest(DataClassJsonMixin):
    """
    https://gitlab.com/kicad/code/kicad/-/raw/master/api/schemas/api.v1.schema.json
    """

    @dataclass_json(letter_case=LetterCase.KEBAB)  # type: ignore
    @dataclass
    class Runtime(DataClassJsonMixin):
        type: Literal["python", "exec"]
        # min_version: str | None = None

    @dataclass_json(letter_case=LetterCase.KEBAB)  # type: ignore
    @dataclass
    class Action(DataClassJsonMixin):
        identifier: str
        name: str
        description: str
        show_button: bool
        scopes: list[
            Literal["pcb", "schematic", "footprint", "symbol", "project_manager"]
        ]
        entrypoint: str
        icons_light: list[str]
        icons_dark: list[str]

    identifier: str
    name: str
    description: str
    runtime: Runtime
    actions: list[Action]


@dataclass
class Command:
    command: list[str]
    kicad_action: KicadPluginManifest.Action
    icon_path: Path


def install_kicad_ipc_plugin() -> None:
    """Install the kicad plugin."""
    # Get the current Python interpreter path
    import sys

    python_path = sys.executable

    COMMANDS = [
        Command(
            command=["layout-sync"],
            kicad_action=KicadPluginManifest.Action(
                identifier="layout-sync",
                name="Layout Sync",
                description="Layout components on PCB in same spatial relationships as "
                "components on sub layouts. Warning: this will save the PCB file.",
                show_button=True,
                scopes=["pcb"],
                entrypoint="layout-sync.py",
                icons_light=[],
                icons_dark=[],
            ),
            icon_path=RESOURCE_DIR / "download.png",
        )
    ]

    manifest = KicadPluginManifest(
        identifier="atopile.kicad.plugin",
        name="atopile KiCad Plugin",
        description="KiCad plugin for atopile PCB design automation,"
        " providing layout synchronization and component placement tools.",
        runtime=KicadPluginManifest.Runtime(type="python"),
        actions=[cmd.kicad_action for cmd in COMMANDS],
    )

    for cmd in COMMANDS:
        icon_path = cmd.icon_path.expanduser().resolve().absolute()
        cmd.kicad_action.icons_light.append(str(icon_path))
        cmd.kicad_action.icons_dark.append(str(icon_path))

    base_cmd = [python_path, "-m", "atopile", "kicad-ipc"]

    manifest_str = manifest.to_json(indent=2)

    def _write_plugin(path: Path):
        # Create the directory if it doesn't exist
        path.mkdir(parents=True, exist_ok=True)

        (path / "requirements.txt").write_text("", encoding="utf-8")
        (path / "plugin.json").write_text(manifest_str, encoding="utf-8")

        for cmd in COMMANDS:
            exec_cmd = base_cmd + cmd.command
            command_loader = dedent(f"""
               #!/usr/bin/env python3
               import subprocess

               subprocess.run({exec_cmd!r})
               """)
            (path / cmd.kicad_action.entrypoint).write_text(
                command_loader, encoding="utf-8"
            )

    try:
        plugin_paths = get_plugin_paths()
    except FileNotFoundError:
        raise Exception("KiCAD config path not found. Couldn't install plugin!")

    for plugin_dir in plugin_paths:
        _write_plugin(plugin_dir / "atopile")


def install_kicad_legacy_plugin() -> None:
    """Install the kicad plugin."""
    # Get the current Python interpreter path
    import sys

    python_path = sys.executable

    # Find the path to kicad's plugin directory
    plugin_loader = dedent(f"""
        plugin_path = r"{Path(__file__).parent}"
        atopile_python = r"{python_path}"
        import sys
        import os
        import importlib

        # Store the Python interpreter path for subprocess calls
        os.environ["ATOPILE_PYTHON"] = atopile_python

        if plugin_path not in sys.path:
            sys.path.append(plugin_path)

        # if kicad_plugin is already in sys.modules, reload it
        for module in sys.modules:
            if "legacy" in module:
                importlib.reload(sys.modules[module])

        from legacy.plugin import activate
        activate()
        """)

    def _write_plugin(path: Path):
        # Create the directory if it doesn't exist
        path.mkdir(parents=True, exist_ok=True)

        # Write the plugin loader
        plugin_loader_path = path / "atopile.py"

        if not plugin_loader_path.exists():
            logger.info("Writing plugin loader to %s", plugin_loader_path)
        plugin_loader_path.write_text(plugin_loader, encoding="utf-8")

    try:
        plugin_paths = get_plugin_paths(legacy=True)
    except FileNotFoundError:
        raise Exception("KiCAD config path not found. Couldn't install plugin!")

    for plugin_dir in plugin_paths:
        _write_plugin(plugin_dir)



#!/usr/bin/env python3
"""
Tests for the LSP auto-completion functionality
"""

import sys
import tempfile
from contextlib import contextmanager
from pathlib import Path
from textwrap import dedent
from unittest.mock import Mock, patch

import pytest

from faebryk.core.module import Module
from faebryk.core.parameter import Parameter

# Add the src directory to the Python path
sys.path.insert(0, str(Path(__file__).parent.parent / "src"))

import lsprotocol.types as lsp

import faebryk.library._F as F
from atopile.lsp.lsp_server import (
    _extract_field_reference_before_dot,
    _find_field_reference_node,
    _get_node_completions,
    on_document_completion,
)
from faebryk.core.node import Node


@contextmanager
def mock_file(
    content: str,
):
    with tempfile.NamedTemporaryFile(delete=False) as temp_file:
        temp_file.write(content.encode("utf-8"))
        yield Path(temp_file.name)


class TestFieldReferenceExtraction:
    """Test field reference extraction from cursor position"""

    @pytest.mark.parametrize(
        "line,position,expected",
        [
            # Basic cases
            ("resistor.", 9, "resistor"),
            ("mymodule.instance.", 18, "mymodule.instance"),
            # Array indexing
            ("app.modules[0].", 15, "app.modules[0]"),
            ("resistors[42].", 14, "resistors[42]"),
            ("complex.path[0].sub[1].", 23, "complex.path[0].sub[1]"),
            # Edge cases
            ("test", 4, None),  # No dot
            (".", 1, None),  # Dot at start
            ("  ", 2, None),  # No valid field ref
            ("", 0, None),  # Empty string
            ("a.", 2, "a"),  # Single character
            # Nested brackets
            ("matrix[0][1].", 13, "matrix[0][1]"),
            ("data[key1][key2].", 17, "data[key1][key2]"),
            # Mixed alphanumeric and underscores
            ("my_module.sub_item.", 19, "my_module.sub_item"),
            ("item123.value456.", 17, "item123.value456"),
        ],
    )
    def test_extract_field_reference_before_dot(self, line, position, expected):
        """Test extraction of field references before dots"""
        result = _extract_field_reference_before_dot(line, position)
        assert result == expected, f"Failed for '{line}' at pos {position}"

    def test_extract_field_reference_complex_cases(self):
        """Test more complex field reference extraction scenarios"""
        # Test with whitespace
        line = "  resistor.  "
        result = _extract_field_reference_before_dot(line, 11)
        assert result == "resistor"

        # Test right after dot vs typing after dot
        line = "resistor."
        result = _extract_field_reference_before_dot(line, 9)  # Right after dot
        assert result == "resistor"

        line = "resistor.r"
        result = _extract_field_reference_before_dot(line, 10)  # Typing after dot
        assert result == "resistor"

        # Test with brackets and special characters
        line = "app.modules[hello_world]."
        result = _extract_field_reference_before_dot(line, 25)
        assert result == "app.modules[hello_world]"


class TestNodeCompletions:
    """Test node completion extraction"""

    @pytest.mark.parametrize(
        "node_class, params",
        [
            (
                F.Resistor,
                {
                    "resistance": lsp.CompletionItemKind.Unit,
                    "unnamed[0]": lsp.CompletionItemKind.Interface,
                    "unnamed[1]": lsp.CompletionItemKind.Interface,
                    "max_power": lsp.CompletionItemKind.Unit,
                    "max_voltage": lsp.CompletionItemKind.Unit,
                },
            ),
            (
                F.LEDIndicator,
                {
                    "logic_in": lsp.CompletionItemKind.Interface,
                    "power_in": lsp.CompletionItemKind.Interface,
                    "led": lsp.CompletionItemKind.Field,
                },
            ),
        ],
    )
    def test_get_node_completions(self, node_class, params):
        """Test completion extraction from a Resistor node"""
        node = node_class()
        completions = _get_node_completions(node)

        # Should have at least some completions
        assert len(completions) > 0

        # All completions should be CompletionItem objects
        for item in completions:
            assert isinstance(item, lsp.CompletionItem)
            assert item.label  # Should have a label
            assert item.kind  # Should have a kind
            assert item.detail  # Should have detail text

        # Check for expected resistor attributes
        items = {item.label: item.kind for item in completions}
        assert items == params

    def test_get_node_completions_empty_node(self):
        """Test completion extraction from a minimal node"""

        class EmptyNode(Node):
            pass

        node = EmptyNode()
        completions = _get_node_completions(node)

        # Even empty nodes might have some basic completions
        assert isinstance(completions, list)

        # All items should be valid CompletionItems
        for item in completions:
            assert isinstance(item, lsp.CompletionItem)
            assert item.label

    def test_completion_item_kinds(self):
        """Test that completion items have appropriate kinds"""
        resistor = F.Resistor()
        completions = _get_node_completions(resistor)

        # Check that we have different kinds of completions
        kinds = {item.kind for item in completions}

        # Should have at least some recognizable kinds
        expected_kinds = {
            lsp.CompletionItemKind.Module,
            lsp.CompletionItemKind.Interface,
            lsp.CompletionItemKind.Property,
            lsp.CompletionItemKind.Class,
        }

        # At least some expected kinds should be present
        assert len(kinds.intersection(expected_kinds)) > 0


class TestFieldReferenceResolution:
    """Test field reference to node resolution"""

    def test_find_field_reference_node_simple_missing(self):
        """Test finding a node from a simple field reference"""
        # This is harder to test without a full graph setup
        # For now, test that the function handles invalid cases gracefully

        result = _find_field_reference_node("fake_uri", "pass", "nonexistent.field", 0)
        assert result is None, "Should return None for non-existent field references"

    def test_find_field_reference_node_invalid_input(self):
        """Test field reference resolution with invalid inputs"""
        # Test empty field reference
        result = _find_field_reference_node("fake_uri", "pass", "", 0)
        assert result is None

        # Test with invalid characters
        result = _find_field_reference_node("fake_uri", "pass", "invalid..field", 0)
        assert result is None

    def test_find_field_reference_node_nested_present(self):
        """Test finding a node from a nested field reference"""
        ato = dedent("""
            import Resistor             #1
            module TestModule:          #2
                resistor = new Resistor #3
                                        #4
            """)
        with mock_file(ato) as uri:
            result = _find_field_reference_node(str(uri), ato, "resistor.resistance", 4)
        assert result is not None
        assert isinstance(result, Parameter)

    def test_find_field_reference_node_duplicate(self):
        """Test finding a node from a duplicate field reference"""
        ato = dedent("""
                                        #1
            import Resistor             #2
            module TestModule1:         #3
                resistor = new Resistor #4
                                        #5
            module TestModule2:         #6
                resistor = 100mA        #7
                                        #8
            """)
        with mock_file(ato) as uri:
            result = _find_field_reference_node(str(uri), ato, "resistor", 5)
            assert isinstance(result, Module)
            result2 = _find_field_reference_node(str(uri), ato, "resistor", 8)
            assert isinstance(result2, Parameter)


@contextmanager
def _to_mock(code: str, marker="#|#"):
    dedented = dedent(code)

    marker_row = None
    marker_column = None

    # find marker row & column
    for row, line in enumerate(dedented.split("\n")):
        if marker in line:
            marker_row = row
            marker_column = line.index(marker)
            break
    dedented = dedented.replace(marker, "", count=1)

    if marker_row is None or marker_column is None:
        raise ValueError(f"Marker {marker} not found in code")

    mock_document = Mock()
    mock_document.source = dedented
    mock_document.path = "test.ato"

    mock_params = Mock()
    mock_params.text_document.uri = f"test://{mock_document.path}"
    mock_params.position.line = marker_row
    mock_params.position.character = marker_column

    with patch("atopile.lsp.lsp_server.LSP_SERVER") as mock_server:
        mock_server.workspace.get_text_document.return_value = mock_document
        yield mock_params, mock_document


class TestEndToEndCompletion:
    """End-to-end tests with real ato code scenarios"""

    def test_resistor_completion_end_to_end(self):
        """Test completion for 'resistor = new Resistor' followed by 'resistor.'"""
        # Simulate ato code with a resistor instantiation
        ato = """
            import Resistor
            module TestModule:
                resistor = new Resistor
                resistor.#|#
            """

        with _to_mock(ato) as (mock_params, _):
            result = on_document_completion(mock_params)

            # Should return a CompletionList with resistor attributes
            assert isinstance(result, lsp.CompletionList)
            assert len(result.items) > 0

            # Check for expected resistor completions
            labels = [item.label for item in result.items]
            expected_completions = [
                "resistance",
                "max_power",
                "max_voltage",
                "unnamed[0]",
                "unnamed[1]",
            ]

            for expected in expected_completions:
                assert expected in labels, (
                    f"Expected '{expected}' in completions: {labels}"
                )

    def test_led_indicator_completion_end_to_end(self):
        """Test completion for 'led = new LEDIndicator' followed by 'led.'"""
        ato = """
            import LEDIndicator
            module TestModule:
                led = new LEDIndicator
                led.#|#
            """

        with _to_mock(ato) as (mock_params, _):
            result = on_document_completion(mock_params)

            # Should return a CompletionList with LED attributes
            assert isinstance(result, lsp.CompletionList)
            assert len(result.items) > 0

            # Check for expected LED completions
            labels = {item.label for item in result.items}

            assert labels == {"logic_in", "power_in", "led"}

    def test_nested_field_completion_end_to_end(self):
        """Test completion for nested field access like 'module.submodule.field.'"""
        ato = """
            import LEDIndicator
            module TestModule:
                led_indicator = new LEDIndicator
                led_indicator.led.led.#|#
            """

        with _to_mock(ato) as (mock_params, _):
            result = on_document_completion(mock_params)

            # Should return a CompletionList with LED-specific attributes
            assert isinstance(result, lsp.CompletionList)
            labels = {item.label for item in result.items}
            assert labels == {
                "brightness",
                "color",
                "max_brightness",
                "forward_voltage",
                "current",
                "reverse_working_voltage",
                "reverse_leakage_current",
                "max_current",
                "anode",
                "cathode",
            }

    def test_array_access_completion_end_to_end(self):
        """Test completion for array access like 'resistors[0].'"""
        ato = """
            import Resistor
            module TestModule:
                resistors = new Resistor[5]
                resistors[0].#|#
            """

        with _to_mock(ato) as (mock_params, _):
            result = on_document_completion(mock_params)

            # Should return a CompletionList with resistor attributes
            assert isinstance(result, lsp.CompletionList)
            assert len(result.items) > 0

            # Check for resistor-specific completions
            labels = {item.label for item in result.items}
            expected_completions = {
                "resistance",
                "max_power",
                "max_voltage",
                "unnamed[0]",
                "unnamed[1]",
            }

            assert labels == expected_completions

    def test_no_completion_for_invalid_field_reference(self):
        """Test that no completions are provided for invalid field references"""
        ato = """
            import Resistor
            module TestModule:
                resistor = new Resistor
                nonexistent.#|#
            """

        with _to_mock(ato) as (mock_params, _):
            result = on_document_completion(mock_params)

            # Should return None for non-existent field references
            assert result is None

    def test_rhs_completion_end_to_end(self):
        """Test completion for 'resistor = new Resistor' followed by 'resistor.'"""
        # Simulate ato code with a resistor instantiation
        ato = """
            import ElectricPower
            import Resistor
            module TestModule:
                resistor = new Resistor
                power = new ElectricPower
                resistor.unnamed[0] ~ power.#|#
            """

        with _to_mock(ato) as (mock_params, _):
            result = on_document_completion(mock_params)

            # Should return a CompletionList with resistor attributes
            assert isinstance(result, lsp.CompletionList)
            assert len(result.items) > 0

            # Check for expected resistor completions
            labels = [item.label for item in result.items]
            expected_completions = [
                "hv",
                "lv",
                "voltage",
                "max_current",
                "bus_max_current_consumption_sum",
            ]

            for expected in expected_completions:
                assert expected in labels, (
                    f"Expected '{expected}' in completions: {labels}"
                )

    def test_new_keyword_completion_end_to_end(self):
        """Test completion after 'new' keyword with partial type name"""
        # Mock ato content with imports and local definitions
        ato_content = """
            import Resistor
            import Capacitor
            import LED

            module TestModule:
                x = new #|#
                pass

            interface TestInterface:
                pass
        """

        with _to_mock(ato_content) as (mock_params, _):
            types = on_document_completion(mock_params)

            assert isinstance(types, lsp.CompletionList)
            assert len(types.items) > 0

            labels = {item.label for item in types.items}

            # Should find local definitions at minimum
            expected_local_types = {
                "TestModule",
                "TestInterface",
                "Resistor",
                "Capacitor",
                "LED",
            }

            assert expected_local_types.intersection(labels) == expected_local_types

    def test_import_completion_end_to_end(self):
        """Test completion for 'import' keyword"""
        ato = """
            import #|#
            """
        with _to_mock(ato) as (mock_params, _):
            result = on_document_completion(mock_params)

            assert isinstance(result, lsp.CompletionList)
            assert len(result.items) > 0

            labels = {item.label for item in result.items}
            must_contain = {"Resistor", "Capacitor", "LED", "ElectricPower"}
            assert labels.intersection(must_contain) == must_contain

            # TODO check no raw traits (only impl) in list


if __name__ == "__main__":
    pytest.main([__file__, "-v"])



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

from enum import Enum, auto

import faebryk.library._F as F
from faebryk.core.module import Module
from faebryk.libs.library import L
from faebryk.libs.units import P


class MOSFET(Module):
    class ChannelType(Enum):
        N_CHANNEL = auto()
        P_CHANNEL = auto()

    class SaturationType(Enum):
        ENHANCEMENT = auto()
        DEPLETION = auto()

    channel_type = L.p_field(domain=L.Domains.ENUM(ChannelType))
    saturation_type = L.p_field(domain=L.Domains.ENUM(SaturationType))
    gate_source_threshold_voltage = L.p_field(units=P.V)
    max_drain_source_voltage = L.p_field(units=P.V)
    max_continuous_drain_current = L.p_field(units=P.A)
    on_resistance = L.p_field(units=P.ohm)

    source: F.Electrical
    gate: F.Electrical
    drain: F.Electrical

    designator_prefix = L.f_field(F.has_designator_prefix)(
        F.has_designator_prefix.Prefix.Q
    )

    # @L.rt_field
    # def pickable(self) -> F.is_pickable_by_type:
    #     return F.is_pickable_by_type(
    #         F.is_pickable_by_type.Type.MOSFET,
    #         {
    #             "channel_type": self.channel_type,
    #             # TODO: add support in backend
    #             # "saturation_type": self.saturation_type,
    #             "gate_source_threshold_voltage": self.gate_source_threshold_voltage,
    #             "max_drain_source_voltage": self.max_drain_source_voltage,
    #             "max_continuous_drain_current": self.max_continuous_drain_current,
    #             "on_resistance": self.on_resistance,
    #         },
    #     )

    # TODO pretty confusing
    @L.rt_field
    def can_bridge(self):
        return F.can_bridge_defined(in_if=self.source, out_if=self.drain)

    @L.rt_field
    def pin_association_heuristic(self):
        return F.has_pin_association_heuristic_lookup_table(
            mapping={
                self.source: ["S", "Source"],
                self.gate: ["G", "Gate"],
                self.drain: ["D", "Drain"],
            },
            accept_prefix=False,
            case_sensitive=False,
        )

    def __postinit__(self, *args, **kwargs):
        super().__postinit__(*args, **kwargs)
        self.source.add(F.has_net_name("source", level=F.has_net_name.Level.SUGGESTED))
        self.gate.add(F.has_net_name("gate", level=F.has_net_name.Level.SUGGESTED))
        self.drain.add(F.has_net_name("drain", level=F.has_net_name.Level.SUGGESTED))

    usage_example = L.f_field(F.has_usage_example)(
        example="""
        import MOSFET, ElectricLogic, ElectricPower

        mosfet = new MOSFET
        mosfet.channel_type = MOSFET.ChannelType.N_CHANNEL
        mosfet.saturation_type = MOSFET.SaturationType.ENHANCEMENT
        mosfet.gate_source_threshold_voltage = 2.5V +/- 10%
        mosfet.max_drain_source_voltage = 60V
        mosfet.max_continuous_drain_current = 30A
        mosfet.on_resistance = 5mohm +/- 20%
        mosfet.package = "SOT-23"

        # Use as a switch
        gate_control = new ElectricLogic
        power_supply = new ElectricPower
        load = new ElectricLogic

        mosfet.gate ~ gate_control.line
        mosfet.source ~ power_supply.lv
        mosfet.drain ~ load.line
        """,
        language=F.has_usage_example.Language.ato,
    )



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import faebryk.library._F as F
from faebryk.libs.library import L


class NFET(F.MOSFET):
    def __preinit__(self) -> None:
        self.channel_type.alias_is(F.MOSFET.ChannelType.N_CHANNEL)

    usage_example = L.f_field(F.has_usage_example)(
        example="""
        import NFET, ElectricLogic, ElectricPower

        nfet = new NFET
        nfet.gate_source_threshold_voltage = 2.5V +/- 10%
        nfet.max_drain_source_voltage = 60V
        nfet.max_continuous_drain_current = 30A
        nfet.on_resistance = 5mohm +/- 20%
        nfet.package = "SOT-23"

        # Use as low-side switch
        gate_control = new ElectricLogic
        power_supply = new ElectricPower
        load = new ElectricLogic

        nfet.gate ~ gate_control.line
        nfet.source ~ power_supply.lv  # Connect to ground for low-side
        nfet.drain ~ load.line

        # When gate_control is HIGH, NFET conducts (load connected to ground)
        """,
        language=F.has_usage_example.Language.ato,
    )



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import faebryk.library._F as F
from faebryk.core.moduleinterface import ModuleInterface
from faebryk.libs.library import L
from faebryk.libs.units import P


class CAN(ModuleInterface):
    """
    CAN bus interface
    """

    diff_pair: F.DifferentialPair

    speed = L.p_field(units=P.bps)

    def __preinit__(self) -> None:
        self.speed.add(F.is_bus_parameter())

    def __postinit__(self, *args, **kwargs):
        super().__postinit__(*args, **kwargs)
        self.diff_pair.p.line.add(
            F.has_net_name("CAN_H", level=F.has_net_name.Level.SUGGESTED)
        )
        self.diff_pair.n.line.add(
            F.has_net_name("CAN_L", level=F.has_net_name.Level.SUGGESTED)
        )

    usage_example = L.f_field(F.has_usage_example)(
        example="""
        import CAN, ElectricPower, Resistor

        can_bus = new CAN
        can_bus.speed = 250kbps  # Common CAN speeds: 125k, 250k, 500k, 1Mbps
        can_bus.diff_pair.impedance = 120ohm +/- 5%

        # Connect power reference for logic levels
        power_5v = new ElectricPower
        assert power_5v.voltage within 5V +/- 5%
        can_bus.diff_pair.p.reference ~ power_5v
        can_bus.diff_pair.n.reference ~ power_5v

        # Connect to CAN transceiver
        can_transceiver.can_bus ~ can_bus

        # Connect to microcontroller CAN peripheral
        microcontroller.can ~ can_transceiver.mcu_interface

        # CAN termination resistors (120 ohm at each end of bus)
        termination_resistor = new Resistor
        termination_resistor.resistance = 120ohm +/- 1%
        can_bus.diff_pair.p.line ~> termination_resistor ~> can_bus.diff_pair.n.line

        # Common applications: automotive, industrial automation, IoT
        """,
        language=F.has_usage_example.Language.ato,
    )



import logging
import os
import sys
from datetime import datetime
from pathlib import Path

from kipy import KiCad
from kipy.errors import ApiError, ConnectionError

from atopile.cli.logging_ import ALERT
from faebryk.libs.kicad.fileformat_config import C_kicad_config_common
from faebryk.libs.kicad.fileformats_latest import C_kicad_pcb_file
from faebryk.libs.kicad.paths import get_config_common, get_ipc_socket_path
from faebryk.libs.util import (
    compare_dataclasses,
    once,
    round_dataclass,
    sort_dataclass,
    try_or,
)

logger = logging.getLogger(__name__)

# https://gitlab.com/kicad/code/kicad-python
# https://dev-docs.kicad.org/en/apis-and-binding/ipc-api/for-addon-developers/index.html


def running_in_kicad():
    return os.environ.get("KICAD_API_TOKEN") is not None


@once
def enable_plugin_api():
    cfg_file = get_config_common()
    cfg = C_kicad_config_common.from_json(cfg_file.read_text(encoding="utf-8"))
    if not cfg.api.enable_server:
        cfg.api.enable_server = True
        cfg_file.write_text(cfg.to_json(indent=2, ensure_ascii=False), encoding="utf-8")
        logger.warning(f"Enabled plugin api in {cfg_file}. Please restart KiCad.")


def _kicad_socket_files():
    base = get_ipc_socket_path()
    if sys.platform.startswith("win"):
        # named pipe
        pipes = [Path(p) for p in os.listdir(r"\\.\pipe\\")]
        out = [pipe for pipe in pipes if pipe.is_relative_to(base)]
    else:
        # unix socket
        out = list(base.glob("api*.sock"))
    return out


def _get_all_clients():
    socket_files = _kicad_socket_files()
    clients = [
        KiCad(socket_path=f"ipc://{socket_file}", timeout_ms=5000)
        for socket_file in socket_files
    ]
    # try connect
    clients = [
        client
        for client in clients
        if not try_or(client.ping, True, catch=ConnectionError)
    ]
    return clients


def _get_pcbnew_clients():
    clients = _get_all_clients()
    return [pcbnew for client in clients if (pcbnew := PCBnew.from_client(client))]


def _get_pcbnew_client_for_path(pcb_path: Path):
    clients = _get_pcbnew_clients()
    return [pcbnew for pcbnew in clients if pcbnew.matches(pcb_path)]


class PCBnew:
    def __init__(self, client: KiCad):
        self.client = client

    @property
    def board(self):
        return self.client.get_board()

    @property
    def path(self):
        return (
            (
                Path(self.board.document.project.path)
                / self.board.document.board_filename
            )
            .resolve()
            .absolute()
        )

    def has_pending_changes(self, reference: Path | None = None) -> bool:
        # This is hard
        return True

        def _cleanup(raw: str):
            out = C_kicad_pcb_file.loads(raw)
            for fp in out.kicad_pcb.footprints:
                fp.unknown = None

            out = round_dataclass(
                sort_dataclass(
                    out, sort_key=lambda x: getattr(x, "uuid", None) or str(x)
                ),
                6,
            )
            return out

        reference = reference or self.path

        diskstate = _cleanup(
            reference.read_text(encoding="utf-8") if reference.exists() else ""
        )
        memorystate = _cleanup(self.board.get_as_string())

        diff = compare_dataclasses(
            diskstate,
            memorystate,
            skip_keys=("uuid",),
        )
        return bool(diff)

    def reload(self, backup_path: Path | None = None):
        if backup_path:
            now = datetime.now().strftime("%Y%m%d_%H%M%S")
            client_id = (
                Path(self.client._client._socket_path)
                .name.removeprefix("api")
                .removeprefix("-")
                .removesuffix(".sock")
            ) or "0"
            path = backup_path.with_suffix(f".pcbnew-{client_id}.{now}.kicad_pcb")
            logger.log(
                ALERT,
                f"Backing up unsaved pcb changes to {path}",
            )
            self.board.save_as(str(path), overwrite=True, include_project=False)

        self.board.revert()

    @classmethod
    def from_client(cls, client: KiCad):
        try:
            client.get_board()
            return cls(client)
        except ApiError:
            return None

    def matches(self, pcb_path: Path):
        return self.path == pcb_path.expanduser().resolve().absolute()

    @classmethod
    def get_host(cls):
        if not running_in_kicad():
            raise Exception("Not running in KiCad")
        out = cls.from_client(KiCad())
        if not out:
            raise Exception("No PCBnew client found")
        return out


def opened_in_pcbnew(pcb_path: Path | None) -> bool:
    if not pcb_path:
        return bool(_get_pcbnew_clients())
    return bool(_get_pcbnew_client_for_path(pcb_path))


def reload_pcb(pcb_path: Path, backup_path: Path | None = None):
    clients = _get_pcbnew_client_for_path(pcb_path)
    for client in clients:
        client.reload(backup_path=backup_path)



"""CLI command definition for `ato kicad-ipc`."""

import logging
import subprocess
import sys
from pathlib import Path
from typing import Annotated, Optional

import typer

from faebryk.libs.kicad.fileformats_latest import C_kicad_pcb_file

kicad_ipc_app = typer.Typer(rich_markup_mode="rich")

logger = logging.getLogger(__name__)


def _setup_logger():
    from faebryk.libs.kicad.ipc import running_in_kicad
    from faebryk.libs.paths import get_log_file

    if not running_in_kicad():
        return
    formatter = logging.Formatter(
        "%(asctime)s - %(message)s", datefmt="%Y-%m-%d %H:%M:%S"
    )
    file_handler = logging.FileHandler(str(get_log_file("kicad_ipc")), "w", "utf-8")
    file_handler.setFormatter(formatter)
    logger.addHandler(file_handler)


@kicad_ipc_app.command()
# TODO: had to disable because slow and blocking kicad
# @capture("cli:kicad_ipc_layout_sync_start", "cli:kicad_ipc_layout_sync_end")
def layout_sync(
    legacy: Annotated[
        bool,
        typer.Option("--legacy", help="Called from legacy plugin"),
    ] = False,
    board: Annotated[
        Optional[Path],
        typer.Option("--board", help="Legacy only: Path to the PCB file to sync"),
    ] = None,
    include_groups: Annotated[
        Optional[list[str]],
        typer.Option(
            "--include-group",
            help="Legacy only: Specific group(s) to sync. "
            "If not specified, just syncs groups.",
        ),
    ] = None,
    include_fp: Annotated[
        Optional[list[str]],
        typer.Option(
            "--include-fp",
            help="Legacy only: Specific footprint(s) to sync.",
        ),
    ] = None,
    sync_all: Annotated[
        bool,
        typer.Option("--all", help="Debug only: Sync all groups"),
    ] = False,
):
    """
    Synchronize PCB layout groups from their source layouts.

    This command pulls layout information from source PCB files and applies it to
    groups in the target PCB file, preserving relative positions and routing.
    """

    from atopile.config import config as gcfg
    from faebryk.libs.kicad.ipc import PCBnew, reload_pcb
    from faebryk.libs.util import find, not_none, root_by_file

    _setup_logger()
    logger.info("layout_sync")

    if not legacy and (include_groups or include_fp):
        raise typer.BadParameter(
            "setting --include-group or --include-fp is only supported with --legacy"
        )
    if legacy and not board:
        raise typer.BadParameter("must specify --board when using --legacy")

    from faebryk.exporters.pcb.layout.layout_sync import LayoutSync

    if legacy:
        pcb_path = Path(not_none(board))
        if not pcb_path.exists():
            raise FileNotFoundError(f"PCB file not found: {pcb_path}")
    else:
        raise NotImplementedError("KiCAD IPC is too broken to use/test")
        pcbnew = PCBnew.get_host()
        pcb_path = pcbnew.path
        pcbnew.board.save()
        # from kipy.proto.common import KiCadObjectType
        # include_groups = [
        #     g.GetName()
        #     for g in pcbnew.board.get_selection(KiCadObjectType.KOT_PCB_GROUP)
        # ]
        # include_fp = [
        #     fp.m_Uuid.AsString()
        #     for fp in pcbnew.board.get_selection(KiCadObjectType.KOT_PCB_FOOTPRINT)
        # ]

    pcb_path = pcb_path.expanduser().resolve().absolute()
    logger.info(f"pcb_path: {pcb_path}")

    prj_root = root_by_file("ato.yaml", pcb_path.parent)
    gcfg.apply_options(entry=None, working_dir=prj_root)
    build = find(gcfg.project.builds.values(), lambda b: b.paths.layout == pcb_path)
    gcfg.select_build(build.name)

    # Create layout sync instance
    pcb_file = C_kicad_pcb_file.loads(pcb_path)
    sync = LayoutSync(pcb_file.kicad_pcb)

    if not sync.groups:
        logger.warning("No sub layout groups found in pcb")
        return

    # Sync groups
    logger.info("Synchronizing groups...")
    sync.sync_groups()

    group_names = set(include_groups or [])

    # determine groups from fps
    fp_uuids = set(include_fp or [])
    for group in sync.pcb.groups:
        if set(group.members).issubset(fp_uuids):
            group_names.add(not_none(group.name))

    if sync_all:
        group_names = set(sync.groups.keys())

    # Only sync specified groups
    for group_name in group_names:
        if group_name in sync.groups:
            logger.info(f"Pulling layout for group: {group_name}")
            sync.pull_group_layout(group_name)
        else:
            logger.warning(f"Group '{group_name}' not found in layout maps")

    # Save the PCB
    logger.info(f"Saving PCB to {pcb_path}")
    pcb_file.dumps(pcb_path)

    # Reload in KiCad if requested
    logger.info("Reloading PCB in KiCad...")
    if legacy:
        # needs hack because plugin blocks IPC calls
        subprocess.Popen(
            [sys.executable, "-m", "atopile", "kicad-ipc", "reload", pcb_path],
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL,
            stdin=subprocess.DEVNULL,
            start_new_session=True,
        )
    else:
        reload_pcb(pcb_path, backup_path=None)

    logger.info("Layout sync completed successfully")


@kicad_ipc_app.command()
def reload(pcb_path: Path):
    """Reload the PCB in KiCad."""
    from faebryk.libs.kicad.ipc import reload_pcb

    _setup_logger()
    reload_pcb(pcb_path, backup_path=None)



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import time
import unittest
from itertools import combinations
from pathlib import Path

import pytest

from faebryk.libs.util import (
    DAG,
    SharedReference,
    assert_once,
    complete_type_string,
    debounce,
    invert_dict,
    list_match,
    once,
    path_replace,
    times_out,
    zip_non_locked,
)


class TestUtil(unittest.TestCase):
    def test_zip_non_locked(self):
        expected = [(1, 4), (2, 5), (3, 6)]

        for val, ref in zip(zip_non_locked([1, 2, 3], [4, 5, 6]), expected):
            self.assertEqual(val, ref)

        expected = [(1, 4), (2, 5), (2, 6)]
        for val, ref in zip((it := zip_non_locked([1, 2, 3], [4, 5, 6])), expected):
            self.assertEqual(val, ref)
            if val == (2, 5):
                it.advance(1)

    def test_shared_reference(self):
        def all_equal(*args: SharedReference):
            for left, right in combinations(args, 2):
                self.assertIs(left.links, right.links)
                self.assertIs(left(), right())
                self.assertEqual(left, right)

        r1 = SharedReference(1)
        r2 = SharedReference(2)

        self.assertEqual(r1(), 1)
        self.assertEqual(r2(), 2)

        r1.link(r2)

        all_equal(r1, r2)

        r3 = SharedReference(3)
        r3.link(r2)

        all_equal(r1, r2, r3)

        r4 = SharedReference(4)
        r5 = SharedReference(5)

        r4.link(r5)

        all_equal(r4, r5)

        r5.link(r1)

        all_equal(r1, r2, r3, r4, r5)

    def test_once(self):
        global ran
        ran = False

        @once
        def do(val: int):
            global ran
            ran = True
            return val

        self.assertFalse(ran)

        self.assertEqual(do(5), 5)
        self.assertTrue(ran)
        ran = False

        self.assertEqual(do(5), 5)
        self.assertFalse(ran)

        self.assertEqual(do(6), 6)
        self.assertTrue(ran)
        ran = False

        class A:
            @classmethod
            @once
            def do(cls):
                global ran
                ran = True
                return cls

            @once
            def do_inst(self, arg: int):
                global ran
                ran = True
                return arg

        self.assertEqual(A.do(), A)
        self.assertTrue(ran)
        ran = False

        self.assertEqual(A.do(), A)
        self.assertFalse(ran)

        ran = False
        a = A()
        self.assertEqual(a.do_inst(5), 5)
        self.assertTrue(ran)
        ran = False
        self.assertEqual(a.do_inst(5), 5)
        self.assertFalse(ran)

    def test_assert_once(self):
        class A:
            def __init__(self):
                self.a = 1

            @assert_once
            def do(self):
                print("do")
                self.a = 5

            @assert_once
            def do_with_arg(self, arg: int):
                print("do_with_arg", arg)
                self.a = arg

        a = A()
        a.do()
        self.assertEqual(a.a, 5)
        self.assertRaises(AssertionError, a.do)
        a.do_with_arg(3)
        self.assertEqual(a.a, 3)
        self.assertRaises(AssertionError, a.do_with_arg, 2)

    def test_times_out(self):
        @times_out(0.1)
        def do(wait_s: float):
            time.sleep(wait_s)

        self.assertRaises(TimeoutError, do, 0.5)
        do(0)


class _DictTestObjBrokenHash:
    def __init__(self, value: object):
        self.value = value

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, _DictTestObjBrokenHash):
            return False
        return self.value == other.value

    def __hash__(self) -> int:
        return 5


@pytest.mark.parametrize(
    "to_test, expected",
    [
        ({1: "a", 2: "b", 3: "c"}, {"a": [1], "b": [2], "c": [3]}),
        ({1: "a", 2: "a", 3: "a"}, {"a": [1, 2, 3]}),
        ({1: "a", 2: "b", 3: "a"}, {"a": [1, 3], "b": [2]}),
        (
            {
                1: (a := _DictTestObjBrokenHash("a")),
                2: (b := _DictTestObjBrokenHash("b")),
                3: a,
            },
            {a: [1, 3], b: [2]},
        ),
    ],
)
def test_invert_dict(to_test: dict, expected: dict):
    out = invert_dict(to_test)
    assert out == expected


def test_dag():
    dag = DAG[int]()
    dag.add_edge(1, 2)
    dag.add_edge(2, 3)

    assert dag.get(1).children == {2}
    assert dag.get(3).parents == {2}

    dag.add_edge(1, 3)
    assert dag.get(3).parents == {1, 2}
    assert dag.get(2).children == {3}
    assert dag.get(1).children == {2, 3}

    assert not dag.contains_cycles

    dag.add_edge(3, 1)
    assert dag.contains_cycles


def test_dag_topological_sort():
    # Test simple linear dependency
    dag = DAG[str]()
    dag.add_edge("A", "B")
    dag.add_edge("B", "C")

    sorted_nodes = dag.topologically_sorted()
    assert sorted_nodes == ["A", "B", "C"]

    # Test diamond dependency
    dag2 = DAG[str]()
    dag2.add_edge("A", "B")
    dag2.add_edge("A", "C")
    dag2.add_edge("B", "D")
    dag2.add_edge("C", "D")

    sorted_nodes2 = dag2.topologically_sorted()
    # A must come before B and C, B and C must come before D
    assert sorted_nodes2.index("A") < sorted_nodes2.index("B")
    assert sorted_nodes2.index("A") < sorted_nodes2.index("C")
    assert sorted_nodes2.index("B") < sorted_nodes2.index("D")
    assert sorted_nodes2.index("C") < sorted_nodes2.index("D")

    # Test with disconnected components
    dag3 = DAG[int]()
    dag3.add_edge(1, 2)
    dag3.add_edge(3, 4)
    dag3.add_edge(5, 6)

    sorted_nodes3 = dag3.topologically_sorted()
    # Each pair should maintain order
    assert sorted_nodes3.index(1) < sorted_nodes3.index(2)
    assert sorted_nodes3.index(3) < sorted_nodes3.index(4)
    assert sorted_nodes3.index(5) < sorted_nodes3.index(6)

    # Test cycle detection in topological sort
    dag_cycle = DAG[str]()
    dag_cycle.add_edge("A", "B")
    dag_cycle.add_edge("B", "C")
    dag_cycle.add_edge("C", "A")

    with pytest.raises(
        ValueError, match="Cannot topologically sort a graph with cycles"
    ):
        dag_cycle.topologically_sorted()

    # Test empty DAG
    dag_empty = DAG[int]()
    assert dag_empty.topologically_sorted() == []

    # Test single node (no edges)
    dag_single = DAG[str]()
    dag_single.add_or_get("A")
    assert dag_single.topologically_sorted() == ["A"]


def test_dag_get_subgraph():
    # Create a DAG with multiple connected components
    dag = DAG[str]()

    # Component 1: A -> B -> C -> D
    dag.add_edge("A", "B")
    dag.add_edge("B", "C")
    dag.add_edge("C", "D")

    # Component 2: E -> F -> G
    dag.add_edge("E", "F")
    dag.add_edge("F", "G")

    # Component 3: H -> I (disconnected)
    dag.add_edge("H", "I")

    # Test selecting node D includes all its dependencies
    subgraph1 = dag.get_subgraph(selector_func=lambda x: x == "D")
    sorted1 = subgraph1.topologically_sorted()
    assert sorted1 == ["A", "B", "C", "D"]

    # Test selecting multiple nodes
    subgraph3 = dag.get_subgraph(selector_func=lambda x: x in ["C", "G"])
    sorted3 = subgraph3.topologically_sorted()
    # Should have both chains up to C and G
    assert "A" in sorted3 and "B" in sorted3 and "C" in sorted3
    assert "E" in sorted3 and "F" in sorted3 and "G" in sorted3
    assert "D" not in sorted3  # D is a child of C, not a parent
    assert "H" not in sorted3 and "I" not in sorted3  # Disconnected component

    # Test empty selector
    subgraph4 = dag.get_subgraph(selector_func=lambda x: False)
    assert subgraph4.topologically_sorted() == []

    # Test selector that matches all
    subgraph5 = dag.get_subgraph(selector_func=lambda x: True)
    assert len(subgraph5.topologically_sorted()) == 9  # All nodes

    # Test non-direct dependencies are included
    # Create a longer chain: A -> B -> C -> D -> E
    dag_chain = DAG[str]()
    dag_chain.add_edge("A", "B")
    dag_chain.add_edge("B", "C")
    dag_chain.add_edge("C", "D")
    dag_chain.add_edge("D", "E")

    # Select E should include all ancestors A, B, C, D
    subgraph_chain = dag_chain.get_subgraph(selector_func=lambda x: x == "E")
    sorted_chain = subgraph_chain.topologically_sorted()
    assert sorted_chain == ["A", "B", "C", "D", "E"]

    # Test complex non-direct dependencies
    # Diamond with extension: A -> B -> D -> E
    #                         A -> C -> D -> E
    dag_complex = DAG[str]()
    dag_complex.add_edge("A", "B")
    dag_complex.add_edge("A", "C")
    dag_complex.add_edge("B", "D")
    dag_complex.add_edge("C", "D")
    dag_complex.add_edge("D", "E")

    # Select E should include all ancestors
    subgraph_complex = dag_complex.get_subgraph(selector_func=lambda x: x == "E")
    sorted_complex = subgraph_complex.topologically_sorted()
    assert "A" in sorted_complex
    assert "B" in sorted_complex
    assert "C" in sorted_complex
    assert "D" in sorted_complex
    assert "E" in sorted_complex
    assert len(sorted_complex) == 5


def test_complete_type_string():
    a = {"a": 1, 5: object(), "c": {"a": 1}}
    assert complete_type_string(a) == "dict[str | int, int | object | dict[str, int]]"


def test_ordered_set():
    from ordered_set import OrderedSet

    s = OrderedSet([1, 4, 2, 3, 4])
    assert s == {1: None, 4: None, 2: None, 3: None}
    assert s == {1, 4, 2, 3}
    assert list(s) == [1, 4, 2, 3]

    x = s | {5, 7, 6}
    assert x == {1, 4, 2, 3, 5, 6, 7}
    assert list(x) == [1, 4, 2, 3, 5, 6, 7]


@pytest.mark.parametrize(
    "base, match, expected",
    [
        ([1, 2, 3], [2, 3], [1]),
        ([1, 2, 3], [2, 3, 4], []),
        ([1, 2, 3], [1, 2, 3], [0]),
        ([1, 2, 3], [1, 2, 3, 4], []),
        ([1, 2, 3], [1, 2, 3, 4, 5], []),
    ],
)
def test_list_match(base, match, expected):
    assert list(list_match(base, match)) == expected


@pytest.mark.parametrize(
    "base, match, replacement, expected",
    [
        (Path("a/b/c"), Path("b"), Path("d"), Path("a/d/c")),
        (Path("a/b/c"), Path("b/c"), Path("d"), Path("a/d")),
        (Path("a/c/c"), Path("b/c"), Path("d"), Path("a/c/c")),
        (Path("a/c/c"), Path("c"), Path("d"), Path("a/d/d")),
        (Path("a/c/c/c/c"), Path("c/c"), Path("d"), Path("a/d/d")),
    ],
)
def test_path_replace(base, match, replacement, expected):
    assert path_replace(base, match, replacement) == expected


def test_debounce():
    calls = []
    outer = []

    DEBOUNCE = 0.5

    @debounce(DEBOUNCE)
    def do():
        now = time.time()
        calls.append(now)
        return now

    outer.append(do())
    outer.append(do())
    outer.append(do())
    outer.append(do())

    time.sleep(DEBOUNCE * 3)
    assert len(calls) == 2

    diffs = [calls[i] - calls[i - 1] for i in range(1, len(calls))]

    assert all(diff >= DEBOUNCE for diff in diffs)
    assert len(set(outer)) == 1

    # Rerun
    outer.append(do())
    outer.append(do())
    outer.append(do())

    time.sleep(DEBOUNCE * 2)
    assert len(calls) == 4

    diffs2 = [calls[i] - calls[i - 1] for i in range(1, len(calls))]
    assert all(diff >= DEBOUNCE for diff in diffs2)
    assert len(set(outer)) == 2



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import faebryk.library._F as F
from faebryk.core.moduleinterface import ModuleInterface
from faebryk.libs.library import L


class SPI(ModuleInterface):
    sclk: F.ElectricLogic
    miso: F.ElectricLogic
    mosi: F.ElectricLogic

    @L.rt_field
    def single_electric_reference(self):
        return F.has_single_electric_reference_defined(
            F.ElectricLogic.connect_all_module_references(self)
        )

    def __postinit__(self, *args, **kwargs):
        super().__postinit__(*args, **kwargs)
        self.sclk.line.add(F.has_net_name("SCLK", level=F.has_net_name.Level.SUGGESTED))
        self.miso.line.add(F.has_net_name("MISO", level=F.has_net_name.Level.SUGGESTED))
        self.mosi.line.add(F.has_net_name("MOSI", level=F.has_net_name.Level.SUGGESTED))

    usage_example = L.f_field(F.has_usage_example)(
        example="""
        import SPI, ElectricPower, ElectricLogic

        spi_bus = new SPI

        # Connect power reference for logic levels
        power_3v3 = new ElectricPower
        assert power_3v3.voltage within 3.3V +/- 5%
        spi_bus.sclk.reference ~ power_3v3
        spi_bus.miso.reference ~ power_3v3
        spi_bus.mosi.reference ~ power_3v3

        # Connect to microcontroller
        microcontroller.spi ~ spi_bus

        # Connect to SPI device with chip select
        chip_select = new ElectricLogic
        chip_select.reference ~ power_3v3
        flash_memory.spi ~ spi_bus
        flash_memory.cs ~ chip_select
        """,
        language=F.has_usage_example.Language.ato,
    )



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

from enum import Enum, auto

import faebryk.library._F as F
from faebryk.core.module import Module
from faebryk.libs.library import L
from faebryk.libs.units import P, quantity


class Comparator(Module):
    class OutputType(Enum):
        Differential = auto()
        PushPull = auto()
        OpenDrain = auto()

    common_mode_rejection_ratio = L.p_field(
        units=P.dB,
        likely_constrained=True,
        soft_set=L.Range(quantity(60, P.dB), quantity(120, P.dB)),
        tolerance_guess=10 * P.percent,
    )
    input_bias_current = L.p_field(
        units=P.A,
        likely_constrained=True,
        soft_set=L.Range(1 * P.pA, 1 * P.µA),
        tolerance_guess=20 * P.percent,
    )
    input_hysteresis_voltage = L.p_field(
        units=P.V,
        likely_constrained=True,
        soft_set=L.Range(1 * P.mV, 100 * P.mV),
        tolerance_guess=15 * P.percent,
    )
    input_offset_voltage = L.p_field(
        units=P.V,
        soft_set=L.Range(10 * P.µV, 10 * P.mV),
        tolerance_guess=20 * P.percent,
    )
    propagation_delay = L.p_field(
        units=P.s,
        soft_set=L.Range(10 * P.ns, 1 * P.ms),
        tolerance_guess=15 * P.percent,
    )
    output_type = L.p_field(
        domain=L.Domains.ENUM(OutputType),
        likely_constrained=True,
    )

    power: F.ElectricPower
    inverting_input: F.Electrical
    non_inverting_input: F.Electrical
    output: F.Electrical

    @L.rt_field
    def simple_value_representation(self):
        S = F.has_simple_value_representation_based_on_params_chain.Spec
        return F.has_simple_value_representation_based_on_params_chain(
            S(self.common_mode_rejection_ratio, suffix="CMRR"),
            S(self.input_bias_current, suffix="Ib"),
            S(self.input_hysteresis_voltage, suffix="Vhys"),
            S(self.input_offset_voltage, suffix="Vos"),
            S(self.propagation_delay, suffix="tpd"),
        )

    designator_prefix = L.f_field(F.has_designator_prefix)(
        F.has_designator_prefix.Prefix.U
    )

    usage_example = L.f_field(F.has_usage_example)(
        example="""
        import Comparator, Resistor, ElectricPower, Electrical

        comparator = new Comparator
        comparator.common_mode_rejection_ratio = 80dB +/- 10%
        comparator.input_bias_current = 1nA +/- 50%
        comparator.input_hysteresis_voltage = 5mV +/- 20%
        comparator.input_offset_voltage = 1mV +/- 30%
        comparator.propagation_delay = 100ns +/- 20%
        comparator.output_type = Comparator.OutputType.PushPull
        comparator.package = "SOIC-8"

        # Power supply connections (dual supply)
        power_pos = new ElectricPower
        power_neg = new ElectricPower
        assert power_pos.voltage within 5V +/- 5%
        assert power_neg.voltage within -5V +/- 5%
        comparator.power.hv ~ power_pos.hv
        comparator.power.lv ~ power_neg.lv

        # Create voltage reference with resistor divider
        ref_resistor_high = new Resistor
        ref_resistor_low = new Resistor
        ref_resistor_high.resistance = 10kohm +/- 1%
        ref_resistor_low.resistance = 10kohm +/- 1%

        # Reference voltage = Vcc/2
        power_pos.hv ~> ref_resistor_high ~> comparator.non_inverting_input
        comparator.non_inverting_input ~> ref_resistor_low ~> power_neg.lv

        # Connect input signal to inverting input
        input_signal ~ comparator.inverting_input
        output_signal ~ comparator.output

        # Output will be HIGH when input_signal > reference_voltage
        """,
        language=F.has_usage_example.Language.ato,
    )



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import logging
from dataclasses import dataclass, field
from typing import Any, Optional

import pytest
from dataclasses_json import CatchAll

import faebryk.library._F as F  # noqa: F401  # This is required to prevent a circular import
from faebryk.libs.kicad.fileformats_latest import C_kicad_pcb_file
from faebryk.libs.sexp.dataclass_sexp import (
    DecodeError,
    SymEnum,
    dumps,
    filter_fields,
    loads,
    sexp_field,
    visit_dataclass,
)
from faebryk.libs.test.fileformats import (
    _FPLIB_DIR,  # noqa: F401
    _NETLIST_DIR,  # noqa: F401
    _PCB_DIR,  # noqa: F401
    _PRJ_DIR,  # noqa: F401
    _SCH_DIR,  # noqa: F401
    _SYM_DIR,  # noqa: F401
    _VERSION_DIR,  # noqa: F401
    DEFAULT_VERSION,  # noqa: F401
    PCBFILE,
)

logger = logging.getLogger(__name__)


def _unformat(s: str) -> str:
    return s.replace("\n", "").replace(" ", "")


@dataclass
class Container:
    @dataclass(kw_only=True)
    class SomeDataclass:
        class E_Numbers(SymEnum):
            ONE = "one"
            TWO = "two"

        a: int = field(**sexp_field(positional=True))
        b: bool = field(**sexp_field(positional=True))
        c: Optional[bool] = field(**sexp_field(positional=True), default=None)
        d: Optional[E_Numbers] = field(**sexp_field(positional=True), default=None)
        f: int
        g: bool
        h: Optional[bool] = None
        i: Optional[E_Numbers] = None

    some_dataclass: SomeDataclass


@pytest.mark.parametrize(
    "container,str_sexp",
    [
        (
            Container(
                some_dataclass=Container.SomeDataclass(
                    a=1, b=True, c=None, d=None, f=2, g=False, h=None, i=None
                )
            ),
            "(some_dataclass 1 yes (f 2) (g no))",
        ),
        (
            Container(
                some_dataclass=Container.SomeDataclass(
                    a=1,
                    b=True,
                    c=False,
                    d=Container.SomeDataclass.E_Numbers.ONE,
                    f=2,
                    g=False,
                    h=True,
                    i=Container.SomeDataclass.E_Numbers.TWO,
                )
            ),
            "(some_dataclass 1 yes no one (f 2) (g no) (h yes) (i two))",
        ),
    ],
)
def test_no_unknowns(container, str_sexp):
    cereal = dumps(container)

    assert _unformat(cereal) == _unformat(str_sexp)

    assert loads(cereal, Container) == container


def test_extra_positional():
    with pytest.raises(DecodeError):
        loads('(some_dataclass 1 yes no "one" gibberish (e 2) (f no))', Container)


def test_empty_unknowns():
    @dataclass
    class Container:
        @dataclass
        class SomeDataclass:
            a: int
            b: str
            unknown: CatchAll = None

        some_dataclass: SomeDataclass

    container = Container(some_dataclass=Container.SomeDataclass(a=1, b="hello"))

    cereal = dumps(container)

    assert _unformat(cereal) == _unformat('(some_dataclass (a 1) (b "hello"))')

    assert loads(cereal, Container) == container


def test_unknowns():
    @dataclass
    class Container:
        @dataclass
        class SomeDataclass:
            a: int
            b: str
            unknown: CatchAll = None

        some_dataclass: SomeDataclass

    cereal = (
        '(some_dataclass gibberish (thingo) (a 1) (b "hello") gibberish (thingo) '
        '(random_key "random_value") (whats_this (who_even_knows True)))'
    )

    container = loads(cereal, Container)

    assert container.some_dataclass.a == 1
    assert container.some_dataclass.b == "hello"
    assert container.some_dataclass.unknown  # there is content

    assert _unformat(dumps(container)) == _unformat(cereal)


def test_print_sexp():
    pcb = C_kicad_pcb_file.loads(PCBFILE)
    dfs = list(visit_dataclass(pcb))
    for it in dfs:
        obj = it.value
        path = it.path
        name_path = it.name_path
        name = "".join(name_path)
        logger.debug(f"{name:70} {[type(p).__name__ for p in path + [obj]]}")

    logger.debug("-" * 80)

    level2 = [p for p in dfs if len(p.path) == 2]
    for it in level2:
        obj = it.value
        path = it.path
        name_path = it.name_path
        name = "".join(name_path)
        logger.debug(f"{name:70}  {[type(p).__name__ for p in path + [obj]]}")


def test_filter_fields():
    from faebryk.libs.sexp.dataclass_sexp import filter_fields

    @dataclass
    class Inner:
        keep_me: str
        remove_me: str
        also_keep: int

    @dataclass
    class Outer:
        name: str
        remove_me: str
        inner: Inner
        inner_list: list[Inner]
        inner_dict: dict[str, Inner]

    # Create test data
    inner1 = Inner(keep_me="hello", remove_me="secret1", also_keep=42)
    inner2 = Inner(keep_me="world", remove_me="secret2", also_keep=99)
    inner3 = Inner(keep_me="test", remove_me="secret3", also_keep=7)

    outer = Outer(
        name="test_outer",
        remove_me="top_secret",
        inner=inner1,
        inner_list=[inner2, inner3],
        inner_dict={"key1": inner1, "key2": inner2},
    )

    # Test filtering single field
    filtered = filter_fields(outer, ["remove_me"])

    # Check that remove_me is set to None in the filtered object
    assert filtered.name == "test_outer"
    assert filtered.remove_me is None

    # Check nested objects
    assert filtered.inner.keep_me == "hello"
    assert filtered.inner.remove_me is None
    assert filtered.inner.also_keep == 42

    # Check list of objects
    assert len(filtered.inner_list) == 2
    for item in filtered.inner_list:
        assert item.keep_me in ["world", "test"]
        assert item.remove_me is None
        assert item.also_keep in [99, 7]

    # Check dict of objects
    assert len(filtered.inner_dict) == 2
    for key, item in filtered.inner_dict.items():
        assert item.keep_me in ["hello", "world"]
        assert item.remove_me is None
        assert item.also_keep in [42, 99]


def test_filter_fields_multiple():
    from faebryk.libs.sexp.dataclass_sexp import filter_fields

    @dataclass
    class TestClass:
        field1: str
        field2: int
        field3: float
        field4: bool
        field5: Optional[str] = None

    obj = TestClass(
        field1="keep", field2=42, field3=3.14, field4=True, field5="optional"
    )

    # Filter multiple fields
    filtered = filter_fields(obj, ["field2", "field4"])

    assert filtered.field1 == "keep"
    assert filtered.field2 is None
    assert filtered.field3 == 3.14
    assert filtered.field4 is None
    assert filtered.field5 == "optional"


def test_filter_fields_edge_cases():
    from faebryk.libs.sexp.dataclass_sexp import filter_fields

    @dataclass
    class SimpleClass:
        value: str

    # Test with empty field list
    obj = SimpleClass(value="test")
    filtered = filter_fields(obj, [])
    assert filtered.value == "test"

    # Test with non-existent field
    filtered = filter_fields(obj, ["non_existent"])
    assert filtered.value == "test"

    # Test with dataclass containing list
    @dataclass
    class ContainerWithList:
        items: list[SimpleClass]

    obj_with_list = ContainerWithList(items=[SimpleClass("a"), SimpleClass("b")])
    filtered_with_list = filter_fields(obj_with_list, ["value"])
    assert len(filtered_with_list.items) == 2
    for item in filtered_with_list.items:
        assert item.value is None

    # Test with dataclass containing tuple
    @dataclass
    class ContainerWithTuple:
        items: tuple[SimpleClass, SimpleClass]

    obj_with_tuple = ContainerWithTuple(items=(SimpleClass("x"), SimpleClass("y")))
    filtered_with_tuple = filter_fields(obj_with_tuple, ["value"])
    assert len(filtered_with_tuple.items) == 2
    assert isinstance(filtered_with_tuple.items, tuple)
    for item in filtered_with_tuple.items:
        assert item.value is None


def test_filter_uuids():
    def filter_uuids(obj: Any) -> Any:
        return filter_fields(obj, ["uuid"])

    @dataclass
    class WithUuid:
        name: str
        uuid: str
        data: int

    obj = WithUuid(name="test", uuid="12345-67890", data=42)
    filtered = filter_uuids(obj)

    assert filtered.name == "test"
    assert filtered.uuid is None
    assert filtered.data == 42



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

from typing import Self

import faebryk.library._F as F
from faebryk.core.moduleinterface import ModuleInterface
from faebryk.libs.library import L
from faebryk.libs.units import P


class DifferentialPair(ModuleInterface):
    p: F.ElectricSignal
    n: F.ElectricSignal

    impedance = L.p_field(
        units=P.Ω,
        likely_constrained=True,
        soft_set=L.Range(10 * P.Ω, 100 * P.Ω),
        tolerance_guess=10 * P.percent,
    )

    @L.rt_field
    def single_electric_reference(self):
        return F.has_single_electric_reference_defined(
            F.ElectricSignal.connect_all_module_references(self)
        )

    def terminated(self) -> Self:
        terminated_bus = type(self)()
        rs = terminated_bus.add_to_container(2, F.Resistor)
        for r in rs:
            r.resistance.alias_is(self.impedance)

        terminated_bus.p.line.connect_via(rs[0], self.p.line)
        terminated_bus.n.line.connect_via(rs[1], self.n.line)
        self.connect_shallow(terminated_bus)

        return terminated_bus

    def __postinit__(self, *args, **kwargs):
        """Attach required net name suffixes for KiCad differential pair detection.

        Ensures nets associated with the positive and negative lines end with
        `_p` and `_n` respectively. The naming algorithm will append these
        required affixes while still deconflicting names globally.
        """
        super().__postinit__(*args, **kwargs)
        # Apply suffixes to the electrical lines of the signals
        self.p.line.add(F.has_net_name_affix.suffix("_P"))
        self.n.line.add(F.has_net_name_affix.suffix("_N"))

    usage_example = L.f_field(F.has_usage_example)(
        example="""
        import DifferentialPair, ElectricPower

        diff_pair = new DifferentialPair
        diff_pair.impedance = 100ohm +/- 10%  # Common for high-speed signals

        # Connect power reference for signal levels
        power_3v3 = new ElectricPower
        assert power_3v3.voltage within 3.3V +/- 5%
        diff_pair.p.reference ~ power_3v3
        diff_pair.n.reference ~ power_3v3

        # Connect between transmitter and receiver
        transmitter.diff_out ~ diff_pair
        diff_pair ~ receiver.diff_in

        # For terminated transmission line
        terminated_pair = diff_pair.terminated()
        transmitter.diff_out ~ terminated_pair

        # Common applications: USB, Ethernet, PCIe, HDMI
        usb_dp_dn = new DifferentialPair
        usb_dp_dn.impedance = 90ohm +/- 10%
        """,
        language=F.has_usage_example.Language.ato,
    )



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import os
import shutil
import sys
from datetime import datetime
from pathlib import Path

import platformdirs

from faebryk.libs.util import once


@once
def get_config_dir() -> Path:
    def _cleanup_legacy_dir(legacy_path: Path, out: Path) -> None:
        known_files = [
            "config.yaml",
            "configured_for.yaml",
            "telemetry.yaml",
        ]

        for file in known_files:
            if (legacy_path / file).exists():
                if not (out / file).exists():
                    shutil.move(legacy_path / file, out / file)
                else:
                    (legacy_path / file).unlink()

        try:
            legacy_path.rmdir()
        except OSError:
            pass

    out = None
    try:
        out = Path(platformdirs.user_config_dir("atopile"))
    except Exception:
        pass

    if not out or sys.platform == "darwin":
        out = (
            Path(os.environ.get("XDG_CONFIG_HOME", Path.home() / ".config")) / "atopile"
        )

    out.mkdir(parents=True, exist_ok=True)

    # handle legacy
    if sys.platform in ["linux", "darwin"]:
        # chronological order
        for legacy_path in [Path.home() / ".atopile", Path.home() / "atopile"]:
            _cleanup_legacy_dir(legacy_path, out)

    return out


def get_data_dir() -> Path:
    try:
        return Path(platformdirs.user_data_dir("atopile"))
    except Exception:
        return Path.home() / ".local" / "share" / "atopile"


def get_cache_dir() -> Path:
    try:
        return Path(platformdirs.user_cache_dir("atopile"))
    except Exception:
        return Path.home() / ".cache" / "atopile"


def get_log_dir() -> Path:
    return Path(platformdirs.user_log_dir("atopile", ensure_exists=True))


def get_log_file(name: str) -> Path:
    timestamp = datetime.now().strftime("%Y%m%d_%H%M%S")
    return get_log_dir() / f"{name}_{timestamp}.log"



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import logging

import faebryk.library._F as F
from faebryk.core.module import Module
from faebryk.libs.library import L

logger = logging.getLogger(__name__)


class TestPoint(Module):
    """
    Basic test point.
    """

    contact: F.Electrical

    designator_prefix = L.f_field(F.has_designator_prefix)(
        F.has_designator_prefix.Prefix.TP
    )

    def __preinit__(self):
        self.contact.add(F.requires_external_usage())

    usage_example = L.f_field(F.has_usage_example)(
        example="""
        import TestPoint

        test_point = new TestPoint

        # Connect to signal you want to probe
        signal_to_test ~ test_point.contact
        """,
        language=F.has_usage_example.Language.ato,
    )



from dataclasses import dataclass

from dataclasses_json import CatchAll, DataClassJsonMixin, Undefined, dataclass_json


@dataclass_json(undefined=Undefined.INCLUDE)
@dataclass(kw_only=True)
class C_kicad_config_common(DataClassJsonMixin):
    @dataclass_json(undefined=Undefined.INCLUDE)
    @dataclass(kw_only=True)
    class C_kicad_config_common_api(DataClassJsonMixin):
        enable_server: bool
        interpreter_path: str
        unknown: CatchAll = None

    api: C_kicad_config_common_api
    unknown: CatchAll = None



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import logging

import faebryk.library._F as F
from faebryk.core.module import Module
from faebryk.libs.library import L
from faebryk.libs.units import P
from faebryk.libs.util import times

logger = logging.getLogger(__name__)


class ResistorArray(Module):
    resistance = L.p_field(units=P.ohm)
    rated_power = L.p_field(units=P.W)
    rated_voltage = L.p_field(units=P.V)

    @L.rt_field
    def resistors(self):
        return times(self._resistor_count, F.Resistor)

    designator_prefix = L.f_field(F.has_designator_prefix)(
        F.has_designator_prefix.Prefix.R
    )

    def __init__(self, resistor_count: int = 4):
        super().__init__()
        self._resistor_count = resistor_count

    def __preinit__(self):
        for resistor in self.resistors:
            resistor.resistance = self.resistance
            resistor.max_power = self.rated_power
            resistor.max_voltage = self.rated_voltage

    usage_example = L.f_field(F.has_usage_example)(
        example="""
        import ResistorArray, ElectricPower, ElectricLogic

        # Create 8-resistor array for pull-ups
        pullup_array = new ResistorArray(resistor_count=8)
        pullup_array.resistance = 10kohm +/- 5%
        pullup_array.rated_power = 125mW
        pullup_array.rated_voltage = 50V
        pullup_array.package = "4816"  # 8-pin SIP package

        # Connect power supply
        power_3v3 = new ElectricPower
        assert power_3v3.voltage within 3.3V +/- 5%

        # Use for GPIO pull-ups
        gpio_signals = new ElectricLogic[8]
        for i in range(8):
            gpio_signals[i].reference ~ power_3v3
            gpio_signals[i].line ~> pullup_array.resistors[i] ~> power_3v3.hv
            microcontroller.gpio[i] ~ gpio_signals[i].line

        # Alternative: 4-resistor array for I2C/SPI bus termination
        termination_array = new ResistorArray(resistor_count=4)
        termination_array.resistance = 33ohm +/- 1%
        termination_array.package = "4806"  # 4-pin array

        # Common applications: pull-up/pull-down networks, bus termination,
        # voltage dividers, current limiting, LED drivers
        """,
        language=F.has_usage_example.Language.ato,
    )



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

from typing import TypeGuard

import faebryk.library._F as F
from faebryk.core.module import Module
from faebryk.core.moduleinterface import ModuleInterface
from faebryk.libs.library import L
from faebryk.libs.util import once


class _TSwitch(Module):
    T = ModuleInterface

    def __init__(self, t: type[T]):
        super().__init__()
        self.t = t

    @staticmethod
    def is_instance(obj: Module, t: type[T]) -> bool:
        return isinstance(obj, _TSwitch) and issubclass(obj.t, t)


@once  # This means we can use a normal "isinstance" to test for them
def Switch[T: ModuleInterface](interface_type: type[T]):
    class _Switch(_TSwitch):
        def __init__(self) -> None:
            super().__init__(interface_type)

        designator_prefix = L.f_field(F.has_designator_prefix)(
            F.has_designator_prefix.Prefix.S
        )
        attach_to_footprint: F.can_attach_to_footprint_symmetrically

        unnamed = L.list_field(2, interface_type)

        @L.rt_field
        def can_bridge(self):
            return F.can_bridge_defined(*self.unnamed)

        @staticmethod
        def is_instance(obj: Module) -> "TypeGuard[_Switch]":
            return _TSwitch.is_instance(obj, interface_type)

    return _Switch


# Usage example for the Switch factory
usage_example = L.f_field(F.has_usage_example)(
    example="""
    import Switch, Electrical, ElectricLogic, ElectricPower

    # Create electrical switch (SPST)
    ElectricalSwitch = Switch(Electrical)
    electrical_switch = new ElectricalSwitch

    # Connect switch terminals
    input_signal ~ electrical_switch.unnamed[0]
    output_signal ~ electrical_switch.unnamed[1]

    # Create logic-level switch
    LogicSwitch = Switch(ElectricLogic)
    logic_switch = new LogicSwitch

    # Connect power reference for logic switch
    power_3v3 = new ElectricPower
    logic_switch.unnamed[0].reference ~ power_3v3
    logic_switch.unnamed[1].reference ~ power_3v3

    # Connect logic signals
    logic_input ~ logic_switch.unnamed[0].line
    logic_output ~ logic_switch.unnamed[1].line

    # Common uses: user buttons, reed switches, micro switches
    """,
    language=F.has_usage_example.Language.ato,
)



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import logging

import faebryk.library._F as F
from faebryk.libs.library import L
from faebryk.libs.units import P

logger = logging.getLogger(__name__)


class TVS(F.Diode):
    reverse_breakdown_voltage = L.p_field(units=P.V)

    # @L.rt_field
    # def pickable(self):
    #     return F.is_pickable_by_type(
    #         F.is_pickable_by_type.Type.TVS,
    #         {
    #             "forward_voltage": self.forward_voltage,
    #             "reverse_working_voltage": self.reverse_working_voltage,
    #             "reverse_leakage_current": self.reverse_leakage_current,
    #             "max_current": self.max_current,
    #             "reverse_breakdown_voltage": self.reverse_breakdown_voltage,
    #         },
    #     )

    usage_example = L.f_field(F.has_usage_example)(
        example="""
        import TVS, ElectricPower, Electrical

        tvs = new TVS
        tvs.reverse_breakdown_voltage = 5.1V +/- 5%
        tvs.max_current = 1A
        tvs.reverse_working_voltage = 5V
        tvs.package = "SOD-123"

        # Connect TVS for power line protection
        power_supply = new ElectricPower
        protected_line = new Electrical

        # TVS protects against voltage spikes
        protected_line ~ tvs.anode
        tvs.cathode ~ power_supply.lv  # Connect to ground

        # Bidirectional TVS for signal line protection
        signal_tvs = new TVS
        signal_tvs.reverse_breakdown_voltage = 3.3V +/- 5%
        signal_line = new Electrical

        signal_line ~ signal_tvs.anode
        signal_tvs.cathode ~ power_supply.lv

        # Common applications: ESD protection, power surge protection
        # TVS clamps voltage spikes to protect sensitive components
        """,
        language=F.has_usage_example.Language.ato,
    )



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

from faebryk.core.trait import TraitImpl
from faebryk.libs.library import L


class has_net_name_affix(L.Trait.decless()):
    """Require a fixed prefix/suffix to be applied to a derived net name.

    This does not set or suggest the base name, it enforces that when a name is
    generated it is of the form: `${required_prefix}${base}${required_suffix}`.

    Typical usage is for differential pairs, e.g. enforce `_p` / `_n` suffixes.
    Attach to an `F.Electrical` interface (e.g. a `.line`).
    """

    def __init__(
        self, required_prefix: str | None = None, required_suffix: str | None = None
    ):
        super().__init__()
        self.required_prefix = required_prefix
        self.required_suffix = required_suffix

    @classmethod
    def prefix(cls, value: str) -> "has_net_name_affix":
        return cls(required_prefix=value, required_suffix=None)

    @classmethod
    def suffix(cls, value: str) -> "has_net_name_affix":
        return cls(required_prefix=None, required_suffix=value)

    @classmethod
    def both(cls, prefix: str, suffix: str) -> "has_net_name_affix":
        return cls(required_prefix=prefix, required_suffix=suffix)

    def handle_duplicate(self, old: TraitImpl, node: L.Node) -> bool:
        # If re-added, keep the more specific (non-None) values; error on conflicts
        assert isinstance(old, has_net_name_affix)
        # Merge if compatible
        if (
            self.required_prefix
            and old.required_prefix
            and self.required_prefix != old.required_prefix
        ):
            # Different required prefixes are incompatible; let caller decide
            return super().handle_duplicate(old, node)
        if (
            self.required_suffix
            and old.required_suffix
            and self.required_suffix != old.required_suffix
        ):
            return super().handle_duplicate(old, node)

        # Prefer new non-None values
        if self.required_prefix is not None:
            old.required_prefix = self.required_prefix
        if self.required_suffix is not None:
            old.required_suffix = self.required_suffix
        return False



# Test file for the target dependency system

from unittest.mock import Mock

import pytest

from atopile.targets import Muster, MusterTarget
from faebryk.core.module import Module
from faebryk.core.solver.solver import Solver


def test_muster_basic_dependencies():
    """Test basic dependency registration and sorting."""
    muster = Muster()

    # Create mock functions
    func_a = Mock()
    func_b = Mock()
    func_c = Mock()

    # Register targets with dependencies
    muster.add_target(MusterTarget("A", [], False, func_a, dependencies=[]))
    muster.add_target(
        MusterTarget("B", [], False, func_b, dependencies=[muster.targets["A"]])
    )
    muster.add_target(
        MusterTarget("C", [], False, func_c, dependencies=[muster.targets["B"]])
    )

    # Get all targets by selecting them explicitly
    sorted_targets = muster.select({"A", "B", "C"})
    sorted_names = [t.name for t in sorted_targets]

    assert sorted_names == ["A", "B", "C"]


def test_muster_diamond_dependencies():
    """Test diamond dependency pattern."""
    muster = Muster()

    # Create mock functions
    funcs = {name: Mock() for name in ["A", "B", "C", "D"]}

    # Register targets in diamond pattern
    # A depends on nothing
    # B depends on A
    # C depends on A
    # D depends on B and C
    muster.add_target(MusterTarget("A", [], False, funcs["A"], dependencies=[]))
    muster.add_target(
        MusterTarget("B", [], False, funcs["B"], dependencies=[muster.targets["A"]])
    )
    muster.add_target(
        MusterTarget("C", [], False, funcs["C"], dependencies=[muster.targets["A"]])
    )
    muster.add_target(
        MusterTarget(
            "D",
            [],
            False,
            funcs["D"],
            dependencies=[muster.targets["B"], muster.targets["C"]],
        )
    )

    # Get all targets by selecting them explicitly
    sorted_targets = muster.select({"A", "B", "C", "D"})
    sorted_names = [t.name for t in sorted_targets]

    # A must come first
    assert sorted_names[0] == "A"
    # D must come last
    assert sorted_names[-1] == "D"
    # B and C must come after A but before D
    assert sorted_names.index("B") > sorted_names.index("A")
    assert sorted_names.index("B") < sorted_names.index("D")
    assert sorted_names.index("C") > sorted_names.index("A")
    assert sorted_names.index("C") < sorted_names.index("D")


def test_muster_specific_targets_with_dependencies():
    """Test getting specific targets includes their dependencies."""
    muster = Muster()

    # Create mock functions
    funcs = {name: Mock() for name in ["A", "B", "C", "D", "E"]}

    # Register targets
    muster.add_target(MusterTarget("A", [], False, funcs["A"], dependencies=[]))
    muster.add_target(
        MusterTarget("B", [], False, funcs["B"], dependencies=[muster.targets["A"]])
    )
    muster.add_target(
        MusterTarget("C", [], False, funcs["C"], dependencies=[muster.targets["B"]])
    )
    muster.add_target(MusterTarget("D", [], False, funcs["D"], dependencies=[]))
    muster.add_target(
        MusterTarget("E", [], False, funcs["E"], dependencies=[muster.targets["D"]])
    )

    # Get only C (should include A and B as dependencies)
    sorted_targets = muster.select({"C"})
    sorted_names = [t.name for t in sorted_targets]

    assert sorted_names == ["A", "B", "C"]

    # Get E (should include only D as dependency, not A, B, C)
    sorted_targets = muster.select({"E"})
    sorted_names = [t.name for t in sorted_targets]

    assert sorted_names == ["D", "E"]

    # Test selecting all targets explicitly
    all_targets = muster.select({"A", "B", "C", "D", "E"})
    all_names = [t.name for t in all_targets]
    assert set(all_names) == {"A", "B", "C", "D", "E"}


def test_muster_cycle_detection():
    """Test that cyclic dependencies are detected."""
    muster = Muster()

    # Create mock functions
    funcs = {name: Mock() for name in ["A", "B", "C"]}

    # First register all targets without dependencies
    muster.add_target(MusterTarget("A", [], False, funcs["A"], dependencies=[]))
    muster.add_target(MusterTarget("B", [], False, funcs["B"], dependencies=[]))
    muster.add_target(MusterTarget("C", [], False, funcs["C"], dependencies=[]))

    # Now add the cyclic dependencies directly to the DAG
    muster.dependency_dag.add_edge("A", "B")
    muster.dependency_dag.add_edge("B", "C")
    muster.dependency_dag.add_edge("C", "A")

    # Should raise ValueError due to cycle
    with pytest.raises(
        ValueError, match="Cannot topologically sort a graph with cycles"
    ):
        muster.select({"A", "B", "C"})


def test_muster_register_decorator():
    """Test the register decorator with dependencies."""
    muster = Muster()

    @muster.register("target1")
    def func1(app: Module, solver: Solver) -> None:
        pass

    @muster.register("target2", dependencies=[muster.targets["target1"]])
    def func2(app: Module, solver: Solver) -> None:
        pass

    @muster.register(
        "target3", dependencies=[muster.targets["target1"], muster.targets["target2"]]
    )
    def func3(app: Module, solver: Solver) -> None:
        pass

    # Check targets are registered
    assert "target1" in muster.targets
    assert "target2" in muster.targets
    assert "target3" in muster.targets

    # Check dependencies
    assert muster.targets["target1"].dependencies == []
    assert muster.targets["target2"].dependencies == [muster.targets["target1"]]
    assert muster.targets["target3"].dependencies == [
        muster.targets["target1"],
        muster.targets["target2"],
    ]

    # Check sorting
    sorted_targets = muster.select({"target1", "target2", "target3"})
    sorted_names = [t.name for t in sorted_targets]

    assert sorted_names == ["target1", "target2", "target3"]


def test_muster_disconnected_components():
    """Test that disconnected dependency graphs work correctly."""
    muster = Muster()

    # Create two separate dependency chains
    funcs = {name: Mock() for name in ["A", "B", "C", "D"]}

    # Chain 1: A -> B
    muster.add_target(MusterTarget("A", [], False, funcs["A"], dependencies=[]))
    muster.add_target(
        MusterTarget("B", [], False, funcs["B"], dependencies=[muster.targets["A"]])
    )

    # Chain 2: C -> D (independent of chain 1)
    muster.add_target(MusterTarget("C", [], False, funcs["C"], dependencies=[]))
    muster.add_target(
        MusterTarget("D", [], False, funcs["D"], dependencies=[muster.targets["C"]])
    )

    # Get all sorted targets
    sorted_targets = muster.select({"A", "B", "C", "D"})
    sorted_names = [t.name for t in sorted_targets]

    # Check that dependencies within each chain are respected
    assert sorted_names.index("A") < sorted_names.index("B")
    assert sorted_names.index("C") < sorted_names.index("D")


def test_muster_missing_dependency_error():
    """Test that missing dependencies raise an assertion error."""
    muster = Muster()

    # Add a target that depends on a non-existent target
    func = Mock()
    non_existent_target = MusterTarget("NonExistent", [], False, Mock())

    # Should raise AssertionError for missing dependency
    with pytest.raises(
        AssertionError,
        match="Dependency 'NonExistent' for target 'A' not yet registered",
    ):
        muster.add_target(
            MusterTarget("A", [], False, func, dependencies=[non_existent_target])
        )


def test_muster_non_direct_dependencies():
    """Test that non-direct (transitive) dependencies are included."""
    muster = Muster()

    # Create a longer dependency chain: A -> B -> C -> D -> E
    funcs = {name: Mock() for name in ["A", "B", "C", "D", "E"]}

    muster.add_target(MusterTarget("A", [], False, funcs["A"], dependencies=[]))
    muster.add_target(
        MusterTarget("B", [], False, funcs["B"], dependencies=[muster.targets["A"]])
    )
    muster.add_target(
        MusterTarget("C", [], False, funcs["C"], dependencies=[muster.targets["B"]])
    )
    muster.add_target(
        MusterTarget("D", [], False, funcs["D"], dependencies=[muster.targets["C"]])
    )
    muster.add_target(
        MusterTarget("E", [], False, funcs["E"], dependencies=[muster.targets["D"]])
    )

    # Request only E, should get all ancestors
    sorted_targets = muster.select({"E"})
    sorted_names = [t.name for t in sorted_targets]

    assert sorted_names == ["A", "B", "C", "D", "E"]

    # Test with multiple endpoints that share dependencies
    # Add F that also depends on C
    funcs["F"] = Mock()
    muster.add_target(
        MusterTarget("F", [], False, funcs["F"], dependencies=[muster.targets["C"]])
    )

    # Request both E and F, should get A, B, C once (not duplicated)
    sorted_targets = muster.select({"E", "F"})
    sorted_names = [t.name for t in sorted_targets]

    # Should contain all nodes, with proper ordering
    assert len(sorted_names) == 6  # A, B, C, D, E, F
    assert sorted_names.index("A") < sorted_names.index("B")
    assert sorted_names.index("B") < sorted_names.index("C")
    assert sorted_names.index("C") < sorted_names.index("D")
    assert sorted_names.index("D") < sorted_names.index("E")
    assert sorted_names.index("C") < sorted_names.index("F")



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import faebryk.library._F as F
from faebryk.core.moduleinterface import ModuleInterface
from faebryk.libs.library import L


class SWD(ModuleInterface):
    clk: F.ElectricLogic
    dio: F.ElectricLogic
    swo: F.ElectricLogic
    reset: F.ElectricLogic

    @L.rt_field
    def single_electric_reference(self):
        return F.has_single_electric_reference_defined(
            F.ElectricLogic.connect_all_module_references(self)
        )

    def __postinit__(self, *args, **kwargs):
        super().__postinit__(*args, **kwargs)
        self.clk.line.add(
            F.has_net_name("SWD_CLK", level=F.has_net_name.Level.SUGGESTED)
        )
        self.dio.line.add(
            F.has_net_name("SWD_DIO", level=F.has_net_name.Level.SUGGESTED)
        )
        self.swo.line.add(
            F.has_net_name("SWD_SWO", level=F.has_net_name.Level.SUGGESTED)
        )
        self.reset.line.add(
            F.has_net_name("SWD_RESET", level=F.has_net_name.Level.SUGGESTED)
        )

    usage_example = L.f_field(F.has_usage_example)(
        example="""
        import SWD, ElectricPower, Resistor

        from "x/y/y.ato" import SomeMCU
        from "a/b/c.ato" import SomeDebugger

        swd = new SWD
        microcontroller = new SomeMCU
        debugger = new SomeDebugger

        # Connect power reference for logic levels
        power_3v3 = new ElectricPower
        assert power_3v3.voltage within 3.3V +/- 5%
        swd.reference_shim ~ power_3v3

        # Connect to microcontroller specific pins (mcu has no SWD interface)
        microcontroller.gpio[0] ~ swd.clk
        microcontroller.gpio[1] ~ swd.dio
        microcontroller.gpio[2] ~ swd.swo
        microcontroller.reset ~ swd.reset

        # Connect to debugger (has SWD interface)
        debugger.swd ~ swd

        # Optional pullup resistors for robust operation
        # mostly only on target side, not debugger side
        reset_pullup = new Resistor
        reset_pullup.resistance = 10kohm +/- 5%
        swd.reset.line ~> reset_pullup ~> swd.reset.reference.hv

        # SWD is commonly used for ARM Cortex-M debugging and programming
        """,
        language=F.has_usage_example.Language.ato,
    )



import os
import shutil
import sys
from pathlib import Path

import pytest

from faebryk.libs.util import repo_root as _repo_root
from faebryk.libs.util import run_live

# Get the examples directory relative to this test file
EXAMPLES_DIR = _repo_root() / "examples"


@pytest.mark.parametrize(
    "example",
    [pytest.param(manifest.parent) for manifest in EXAMPLES_DIR.glob("*/ato.yaml")],
    ids=lambda p: p.stem,
)
def test_examples_build(
    example: Path, tmp_path: Path, repo_root: Path, save_tmp_path_on_failure: None
):
    example_copy = tmp_path / example.name
    shutil.copytree(example, example_copy)

    assert example_copy.exists()

    _, stderr, _ = run_live(
        [sys.executable, "-m", "atopile", "build"],
        env={**os.environ, "NONINTERACTIVE": "1"},
        cwd=example_copy,
        stdout=print,
        stderr=print,
    )

    # TODO: add a strict mode to the CLI
    assert "Build successful! 🚀" in stderr
    assert stderr.count("✓") >= 1
    assert stderr.count("✗") == 0

    # expected warnings:
    # - missing kicad-cli for '3d-model' target (in CI only)
    assert stderr.count("⚠") in (0, 1)



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import faebryk.library._F as F
from faebryk.core.module import Module
from faebryk.libs.library import L


class Fan(Module):
    power: F.ElectricPower

    usage_example = L.f_field(F.has_usage_example)(
        example="""
        import Fan, ElectricPower

        fan = new Fan

        # Connect power supply (typically 12V for PC fans)
        power_12v = new ElectricPower
        assert power_12v.voltage within 12V +/- 5%
        assert power_12v.max_current >= 200mA  # Typical fan current

        fan.power ~ power_12v

        # For PWM control, connect through a MOSFET or fan controller
        # fan.power.hv ~ power_12v.hv
        # fan.power.lv ~> pwm_controller ~> power_12v.lv
        """,
        language=F.has_usage_example.Language.ato,
    )



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import faebryk.library._F as F
from faebryk.core.moduleinterface import ModuleInterface
from faebryk.libs.library import L


class UART(ModuleInterface):
    base_uart: F.UART_Base
    rts: F.ElectricLogic
    cts: F.ElectricLogic
    dtr: F.ElectricLogic
    dsr: F.ElectricLogic
    dcd: F.ElectricLogic
    ri: F.ElectricLogic

    # TODO: this creates too many connections in some projects
    # @L.rt_field
    # def single_electric_reference(self):
    #    return F.has_single_electric_reference_defined(
    #       F.ElectricLogic.connect_all_module_references(self)
    #   )

    def __postinit__(self, *args, **kwargs):
        super().__postinit__(*args, **kwargs)
        self.rts.line.add(F.has_net_name("RTS", level=F.has_net_name.Level.SUGGESTED))
        self.cts.line.add(F.has_net_name("CTS", level=F.has_net_name.Level.SUGGESTED))
        self.dtr.line.add(F.has_net_name("DTR", level=F.has_net_name.Level.SUGGESTED))
        self.dsr.line.add(F.has_net_name("DSR", level=F.has_net_name.Level.SUGGESTED))
        self.dcd.line.add(F.has_net_name("DCD", level=F.has_net_name.Level.SUGGESTED))
        self.ri.line.add(F.has_net_name("RI", level=F.has_net_name.Level.SUGGESTED))

    usage_example = L.f_field(F.has_usage_example)(
        example="""
        import UART, ElectricPower

        uart = new UART
        uart.base_uart.baud_rate = 115200

        # Connect power reference for logic levels
        power_3v3 = new ElectricPower
        assert power_3v3.voltage within 3.3V +/- 5%
        uart.base_uart.tx.reference ~ power_3v3
        uart.base_uart.rx.reference ~ power_3v3
        uart.rts.reference ~ power_3v3
        uart.cts.reference ~ power_3v3

        # Connect to microcontroller
        microcontroller.uart ~ uart

        # Connect to external UART device or RS232 transceiver
        external_device.uart_rx ~ uart.base_uart.tx.line
        external_device.uart_tx ~ uart.base_uart.rx.line

        # Hardware flow control (optional)
        external_device.rts ~ uart.cts.line
        external_device.cts ~ uart.rts.line

        # Common baud rates: 9600, 38400, 115200, 230400, 460800
        """,
        language=F.has_usage_example.Language.ato,
    )



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import faebryk.library._F as F
from faebryk.core.moduleinterface import ModuleInterface
from faebryk.libs.library import L


class USB_C(ModuleInterface):
    usb3: F.USB3
    cc1: F.Electrical
    cc2: F.Electrical
    sbu1: F.Electrical
    sbu2: F.Electrical
    rx: F.DifferentialPair
    tx: F.DifferentialPair

    def __postinit__(self, *args, **kwargs):
        super().__postinit__(*args, **kwargs)
        self.cc1.add(F.has_net_name("CC1", level=F.has_net_name.Level.SUGGESTED))
        self.cc2.add(F.has_net_name("CC2", level=F.has_net_name.Level.SUGGESTED))
        self.sbu1.add(F.has_net_name("SBU1", level=F.has_net_name.Level.SUGGESTED))
        self.sbu2.add(F.has_net_name("SBU2", level=F.has_net_name.Level.SUGGESTED))
        self.rx.p.line.add(F.has_net_name("RX", level=F.has_net_name.Level.SUGGESTED))
        self.rx.n.line.add(F.has_net_name("RX", level=F.has_net_name.Level.SUGGESTED))
        self.tx.p.line.add(F.has_net_name("TX", level=F.has_net_name.Level.SUGGESTED))
        self.tx.n.line.add(F.has_net_name("TX", level=F.has_net_name.Level.SUGGESTED))

    usage_example = L.f_field(F.has_usage_example)(
        example="""
        import USB_C, ElectricPower, Resistor

        usb_c = new USB_C

        # Configure differential pair impedances
        usb_c.rx.impedance = 90ohm +/- 10%
        usb_c.tx.impedance = 90ohm +/- 10%
        usb_c.usb3.usb2.dp.impedance = 90ohm +/- 10%
        usb_c.usb3.usb2.dm.impedance = 90ohm +/- 10%

        # Connect power reference for logic levels
        power_3v3 = new ElectricPower
        assert power_3v3.voltage within 3.3V +/- 5%
        usb_c.usb3.usb2.dp.p.reference ~ power_3v3
        usb_c.usb3.usb2.dp.n.reference ~ power_3v3
        usb_c.usb3.usb2.dm.p.reference ~ power_3v3
        usb_c.usb3.usb2.dm.n.reference ~ power_3v3

        # CC resistors for device detection (5.1k for device, 56k for host)
        cc1_resistor = new Resistor
        cc2_resistor = new Resistor
        cc1_resistor.resistance = 5.1kohm +/- 1%  # Device
        cc2_resistor.resistance = 5.1kohm +/- 1%  # Device

        usb_c.cc1 ~> cc1_resistor ~> power_3v3.lv
        usb_c.cc2 ~> cc2_resistor ~> power_3v3.lv

        # Connect to USB controller
        usb_controller.usb_c ~ usb_c
        """,
        language=F.has_usage_example.Language.ato,
    )



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import faebryk.library._F as F
from faebryk.core.moduleinterface import ModuleInterface
from faebryk.libs.library import L
from faebryk.libs.util import times


class MultiSPI(ModuleInterface):
    def __init__(self, data_lane_count: int) -> None:
        super().__init__()
        self._data_lane_count = data_lane_count

    clock: F.ElectricLogic
    chip_select: F.ElectricLogic

    @L.rt_field
    def data(self):
        return times(self._data_lane_count, F.ElectricLogic)

    @L.rt_field
    def single_electric_reference(self):
        return F.has_single_electric_reference_defined(
            F.ElectricLogic.connect_all_module_references(self)
        )

    def __postinit__(self, *args, **kwargs):
        super().__postinit__(*args, **kwargs)
        self.clock.line.add(
            F.has_net_name("clock", level=F.has_net_name.Level.SUGGESTED)
        )
        self.chip_select.line.add(
            F.has_net_name("chip_select", level=F.has_net_name.Level.SUGGESTED)
        )
        for i, line in enumerate(self.data):
            line.add(F.has_net_name(f"data_{i}", level=F.has_net_name.Level.SUGGESTED))

    # ----------------------------------------
    #              usage example
    # ----------------------------------------
    usage_example = L.f_field(F.has_usage_example)(
        example="""
        import MultiSPI, SPI

        # Microcontroller SPI peripheral
        mcu_spi = new SPI

        # Quad-SPI flash interface (4 data lanes)
        qspi = new MultiSPI<int_=4>

        # Connect the buses
        mcu_spi ~ qspi
        """,
        language=F.has_usage_example.Language.ato,
    )



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT


import faebryk.library._F as F
from faebryk.core.module import Module
from faebryk.libs.library import L
from faebryk.libs.units import P


class Crystal_Oscillator(Module):
    # ----------------------------------------
    #     modules, interfaces, parameters
    # ----------------------------------------
    crystal: F.Crystal
    capacitors = L.list_field(2, F.Capacitor)
    current_limiting_resistor: F.Resistor

    xtal_if: F.XtalIF

    # ----------------------------------------
    #               parameters
    # ----------------------------------------
    # https://blog.adafruit.com/2012/01/24/choosing-the-right-crystal-and-caps-for-your-design/
    # http://www.st.com/internet/com/TECHNICAL_RESOURCES/TECHNICAL_LITERATURE/APPLICATION_NOTE/CD00221665.pdf
    _STRAY_CAPACITANCE = L.Range(1 * P.pF, 5 * P.pF)

    @L.rt_field
    def capacitance(self):
        return (self.crystal.load_capacitance - self._STRAY_CAPACITANCE) * 2

    def __preinit__(self):
        for cap in self.capacitors:
            cap.capacitance.constrain_subset(self.capacitance)

        self.current_limiting_resistor.allow_removal_if_zero()
        self.crystal.gnd.connect(self.xtal_if.gnd)
        self.crystal.unnamed[0].connect_via(self.capacitors[0], self.xtal_if.gnd)
        self.crystal.unnamed[1].connect_via(self.capacitors[1], self.xtal_if.gnd)

        self.crystal.unnamed[0].connect_via(
            self.current_limiting_resistor, self.xtal_if.xout
        )
        self.crystal.unnamed[1].connect(self.xtal_if.xin)

    @L.rt_field
    def can_bridge(self):
        return F.can_bridge_defined(self.xtal_if.xin, self.xtal_if.xout)

    usage_example = L.f_field(F.has_usage_example)(
        example="""
        import Crystal_Oscillator, ElectricPower

        crystal_osc = new Crystal_Oscillator

        # Configure crystal parameters
        crystal_osc.crystal.frequency = 16MHz +/- 20ppm
        crystal_osc.crystal.load_capacitance = 18pF +/- 10%
        crystal_osc.crystal.equivalent_series_resistance = 80ohm +/- 20%
        crystal_osc.crystal.package = "HC49U"

        # Connect power for ground reference
        power_3v3 = new ElectricPower
        assert power_3v3.voltage within 3.3V +/- 5%
        crystal_osc.xtal_if.gnd ~ power_3v3.lv

        # Connect to microcontroller crystal pins
        microcontroller.xtal_in ~ crystal_osc.xtal_if.xin
        microcontroller.xtal_out ~ crystal_osc.xtal_if.xout

        # Load capacitors are automatically calculated:
        # C_load = (Crystal_load_cap - Stray_cap) * 2
        # Typically results in 22pF capacitors for 18pF crystal

        # Current limiting resistor prevents overdrive (optional)
        crystal_osc.current_limiting_resistor.resistance = 1kohm +/- 5%

        # Common frequencies: 8MHz, 12MHz, 16MHz, 20MHz, 25MHz
        # Used for: microcontroller clocks, RTC, timing references
        """,
        language=F.has_usage_example.Language.ato,
    )
